from __future__ import annotations

import sys

if sys.version_info >= (3, 11):
    import datetime as dt

    datetime_fromisoformat = dt.datetime.fromisoformat
else:
    import iso8601

    datetime_fromisoformat = iso8601.parse_date



from __future__ import annotations

import os
import sys
from collections.abc import Iterable
from itertools import chain
from pathlib import Path
from shutil import move
from tempfile import mkdtemp


def reveal_folder(path: str | os.PathLike[str]) -> None:
    if sys.platform == 'win32':
        os.startfile(path, 'explore')
    else:
        import click

        click.launch(os.fspath(path), locate=True)


def trash(paths: Iterable[Path], *, dest: Path, missing_ok: bool = False) -> None:
    paths_iter = iter(paths)
    first_path = next(paths_iter, None)

    if first_path is None:
        return

    exc_classes = FileNotFoundError if missing_ok else ()

    parent_folder = mkdtemp(dir=dest, prefix=f'deleted-{first_path.name}-')

    for path in chain((first_path,), paths_iter):
        try:
            move(path, parent_folder)
        except exc_classes:
            pass



from __future__ import annotations

from contextlib import nullcontext

import pytest

from instawow._sources.tukui import TukuiResolver
from instawow.definitions import Defn
from instawow.pkg_models import Pkg
from instawow.results import PkgFilesNotMatching
from instawow.shared_ctx import ConfigBoundCtx
from instawow.wow_installations import Flavour


@pytest.fixture
def tukui_resolver(
    iw_config_ctx: ConfigBoundCtx,
):
    return TukuiResolver(iw_config_ctx.config)


@pytest.mark.parametrize(
    'iw_profile_config_values',
    Flavour,
    indirect=True,
)
@pytest.mark.parametrize('alias', ['tukui', 'elvui'])
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_resolve_addon(
    iw_config_ctx: ConfigBoundCtx,
    tukui_resolver: TukuiResolver,
    alias: str,
):
    defn = Defn('tukui', alias)

    with (
        pytest.raises(PkgFilesNotMatching)
        if (alias == 'tukui' and iw_config_ctx.config.game_flavour is Flavour.Classic)
        or (alias == 'elvui' and iw_config_ctx.config.game_flavour is Flavour.WrathClassic)
        else nullcontext()
    ):
        result = await tukui_resolver.resolve_one(defn, None)

        assert type(result) is Pkg
        assert result.slug == alias


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_changelog_url_format(
    tukui_resolver: TukuiResolver,
):
    defn = Defn('tukui', 'tukui')

    result = await tukui_resolver.resolve_one(defn, None)

    assert result.changelog_url == 'https://api.tukui.org/v1/changelog/tukui#20.41'



from __future__ import annotations

import string
from collections.abc import Callable, Sequence


def tabulate(rows: Sequence[tuple[object, ...]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value: object):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows: Sequence[tuple[str, ...]]):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    norm_rows = [tuple(apply_max_col_width(i) for i in r) for r in rows]
    head, *tail = norm_rows

    base_template = '  '.join(f'{{{{{{0}}{w}}}}}' for w in calc_resultant_col_widths(norm_rows))
    row_template = base_template.format(':<')
    table = '\n'.join(
        (
            base_template.format(':^').format(*head),
            base_template.format('0:-<').format(''),
            *(row_template.format(*r) for r in tail),
        )
    )
    return table


def normalise_names(replace_delim: str) -> Callable[[str], str]:
    trans_table = str.maketrans(dict.fromkeys(string.punctuation, ' '))

    def normalise(value: str):
        return replace_delim.join(value.casefold().translate(trans_table).split())

    return normalise


slugify = normalise_names('-')


def shasum(*values: object) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(map(str, filter(None, values))).encode()).hexdigest()[:32]



from __future__ import annotations

import datetime as dt
import enum
from collections.abc import AsyncIterator, Collection, Sequence
from functools import cached_property
from pathlib import Path
from typing import Any, ClassVar, Protocol, TypedDict, TypeVar

from typing_extensions import NotRequired, deprecated
from yarl import URL

from . import config as _config
from . import pkg_archives, pkg_models
from . import results as R
from .catalogue import cataloguer
from .definitions import Defn, SourceMetadata


class FolderHashCandidate(Protocol):  # pragma: no cover
    @property
    def name(self) -> str: ...
    @property
    def path(self) -> Path: ...


TFolderHashCandidate = TypeVar('TFolderHashCandidate', bound=FolderHashCandidate)


class HeadersIntent(enum.IntEnum):
    Download = enum.auto()


class Resolver(Protocol):  # pragma: no cover
    metadata: ClassVar[SourceMetadata]
    'Static source metadata.'

    requires_access_token: ClassVar[str | None]
    'Access token key or ``None``.'

    archive_opener: ClassVar[pkg_archives.ArchiveOpener | None]
    'Alternative archive opener to use supporting e.g. non-standard archive formats or layouts.'

    def __init__(self, config: _config.ProfileConfig) -> None: ...

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        "Attempt to extract a ``Defn`` alias from a given URL."
        ...

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        "Create headers for resolver HTTP requests."
        ...

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        "Resolve multiple ``Defn``s into packages."
        ...

    async def resolve_one(self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        "Retrieve a changelog from a URI."
        ...

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> list[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        "Find ``Defn``s from folder fingerprint."
        ...

    @classmethod
    def catalogue(cls) -> AsyncIterator[cataloguer.CatalogueEntry]:
        "Enumerate add-ons from the source."
        ...


class PkgCandidate(TypedDict):
    "A subset of the ``Pkg`` constructor's kwargs."

    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: dt.datetime
    version: str
    changelog_url: str
    deps: NotRequired[list[pkg_models.PkgDep]]


class BaseResolver(Resolver, Protocol):
    archive_opener = None

    _config: _config.ProfileConfig

    def __init__(self, config: _config.ProfileConfig) -> None:
        self._config = config

    __orig_init = __init__

    def __init_subclass__(cls) -> None:
        # ``Protocol`` clobbers ``__init__`` on Python < 3.11.
        if cls.__init__ is _DummyResolver.__init__:
            cls.__init__ = cls.__orig_init

    @property
    @deprecated(
        'Use ``self._config.game_flavour`` and ``instawow.shared_ctx.web_client`` instead.'
    )
    def _config_ctx(self) -> Any:
        from types import SimpleNamespace

        from . import shared_ctx

        return SimpleNamespace(
            config=self._config,
            web_client=property(shared_ctx.web_client_var.get),
        )

    @classmethod
    def _get_access_token(
        cls, global_config: _config.GlobalConfig, name: str | None = None
    ) -> str | None:
        name = name or cls.requires_access_token
        if name is not None:
            return getattr(global_config.access_tokens, name, None)

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        return None

    async def make_request_headers(
        self, intent: HeadersIntent | None = None
    ) -> dict[str, str] | None:
        return None

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        from ._progress_reporting import make_incrementing_progress_tracker
        from ._utils.aio import gather

        track_progress = make_incrementing_progress_tracker(
            len(defns), f'Resolving add-ons: {self.metadata.name}'
        )
        results = await gather(
            track_progress(R.resultify_async_exc(self.resolve_one(d, None))) for d in defns
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: Any) -> pkg_models.Pkg:
        extraneous_strategies = defn.strategies.filled.keys() - self.metadata.strategies
        if extraneous_strategies:
            raise R.PkgStrategiesUnsupported(extraneous_strategies)

        pkg_candidate = await self._resolve_one(defn, metadata)
        return pkg_models.Pkg(
            **pkg_candidate,
            source=self.metadata.id,
            options=pkg_models.PkgOptions(**{k: bool(v) for k, v in defn.strategies.items()}),
        )

    async def _resolve_one(self, defn: Defn, metadata: Any) -> PkgCandidate:
        "Resolve a ``Defn`` into a package."
        ...

    async def get_changelog(self, uri: URL) -> str:
        match uri.scheme:
            case 'data' if uri.raw_path.startswith(','):
                import urllib.parse

                return urllib.parse.unquote(uri.raw_path[1:])

            case 'file':
                from ._utils.aio import run_in_thread
                from ._utils.web import file_uri_to_path

                return await run_in_thread(Path(file_uri_to_path(str(uri))).read_text)(
                    encoding='utf-8'
                )

            case 'http' | 'https':
                from . import http, shared_ctx

                async with shared_ctx.web_client.get(
                    uri,
                    expire_after=http.CACHE_INDEFINITELY,
                    headers=await self.make_request_headers(),
                    raise_for_status=True,
                ) as response:
                    return await response.text()

            case _:
                raise ValueError('Unsupported URI with scheme', uri.scheme)

    async def get_folder_hash_matches(
        self, candidates: Collection[TFolderHashCandidate]
    ) -> list[tuple[Defn, frozenset[TFolderHashCandidate]]]:
        return []

    @classmethod
    async def catalogue(cls) -> AsyncIterator[cataloguer.CatalogueEntry]:
        return
        yield


class _DummyResolver(Resolver):
    pass


class Resolvers(dict[str, Resolver]):
    @cached_property
    def archive_opener_dict(self) -> _ResolverArchiveOpenerDict:
        return _ResolverArchiveOpenerDict(self)

    @cached_property
    def priority_dict(self) -> _ResolverPriorityDict:
        return _ResolverPriorityDict(self)

    @cached_property
    def addon_toc_key_and_id_pairs(self) -> Collection[tuple[str, str]]:
        return [
            (r.metadata.addon_toc_key, r.metadata.id)
            for r in self.values()
            if r.metadata.addon_toc_key
        ]


class _ResolverArchiveOpenerDict(dict[str, pkg_archives.ArchiveOpener]):
    def __init__(self, resolvers: Resolvers) -> None:
        super().__init__(
            (r.metadata.id, r.archive_opener) for r in resolvers.values() if r.archive_opener
        )

    def __missing__(self, key: str) -> pkg_archives.ArchiveOpener:
        return pkg_archives.open_zip_archive


class _ResolverPriorityDict(dict[str, float]):
    def __init__(self, resolvers: Resolvers) -> None:
        super().__init__((n, i) for i, n in enumerate(resolvers))

    def __missing__(self, key: str) -> float:
        return float('inf')



from __future__ import annotations

import importlib.resources
from pathlib import Path

import pytest

from instawow.matchers.addon_toc import TocReader


@pytest.fixture
def fake_addon_toc():
    with importlib.resources.as_file(
        importlib.resources.files(__spec__.parent) / 'fixtures' / 'FakeAddon' / 'FakeAddon.toc'
    ) as file:
        yield file


def test_loading_toc_from_path(fake_addon_toc: Path):
    TocReader.from_path(fake_addon_toc)
    with pytest.raises(FileNotFoundError):
        TocReader.from_path(fake_addon_toc.parent / 'MissingAddon.toc')


def test_parsing_toc_entries(fake_addon_toc: Path):
    toc_reader = TocReader.from_path(fake_addon_toc)
    assert dict(toc_reader) == {
        'Normal': 'Normal entry',
        'Compact': 'Compact entry',
    }


def test_toc_entry_indexing(fake_addon_toc: Path):
    toc_reader = TocReader.from_path(fake_addon_toc)
    assert toc_reader['Normal'] == 'Normal entry'
    assert toc_reader['Compact'] == 'Compact entry'
    assert toc_reader.get('Indented') is None
    assert toc_reader.get('Comment') is None
    assert toc_reader.get('Nonexistent') is None


def test_toc_interfaces():
    toc_reader = TocReader('')
    assert toc_reader.interfaces == []

    toc_reader = TocReader('## Interface: 10100')
    assert toc_reader.interfaces == [10100]

    toc_reader = TocReader('## Interface: 10100, 20200')
    assert toc_reader.interfaces == [10100, 20200]

    toc_reader = TocReader('## Interface:  10100 , 20200  ')
    assert toc_reader.interfaces == [10100, 20200]


def test_toc_version():
    toc_reader = TocReader('')
    assert toc_reader.version == ''

    toc_reader = TocReader('## Version: 1')
    assert toc_reader.version == '1'



from __future__ import annotations

import sys
from enum import Enum
from pathlib import Path

import pytest

from instawow.config import ProfileConfig
from instawow.wow_installations import (
    Flavour,
    FlavourVersionRange,
    find_installations,
    get_installation_version_from_addon_dir,
    infer_flavour_from_addon_dir,
)


def test_can_convert_between_flavour_keyed_enum_and_flavour():
    class Foo(Enum):
        Retail = 1
        VanillaClassic = 2
        Classic = 3
        WrathClassic = 4

    assert Flavour.from_flavour_keyed_enum(Foo.Retail) is Flavour.Retail
    assert Flavour.Retail.to_flavour_keyed_enum(Foo) is Foo.Retail


@pytest.mark.parametrize('flavour', list(Flavour))
@pytest.mark.parametrize('affine', [True, False])
def test_flavour_groups_vary_by_flavour_and_affinity(
    flavour: Flavour,
    affine: bool,
):
    flavour_groups = flavour.get_flavour_groups(affine)
    if affine:
        if flavour is Flavour.Classic:
            assert flavour_groups == [(flavour, Flavour.WrathClassic), None]
        else:
            assert flavour_groups == [(flavour,), None]
    else:
        assert flavour_groups == [(flavour,)]


def test_can_extract_flavour_from_version_number():
    assert FlavourVersionRange.from_version(95000) is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version(44000) is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version(34000) is FlavourVersionRange.WrathClassic
    assert FlavourVersionRange.from_version(12300) is FlavourVersionRange.VanillaClassic


def test_can_extract_flavour_from_version_string():
    assert FlavourVersionRange.from_version('9.50.0') is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version('4.4.0') is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version('3.4.0') is FlavourVersionRange.WrathClassic
    assert FlavourVersionRange.from_version('1.23.0') is FlavourVersionRange.VanillaClassic


def test_can_extract_flavour_from_partial_version_string():
    assert FlavourVersionRange.from_version('9.2') is FlavourVersionRange.Retail
    assert FlavourVersionRange.from_version('4.4') is FlavourVersionRange.Classic
    assert FlavourVersionRange.from_version('3.4') is FlavourVersionRange.WrathClassic
    assert FlavourVersionRange.from_version('3') is FlavourVersionRange.Retail


@pytest.mark.parametrize(
    ('path', 'flavour'),
    [
        (
            'wowzerz/_classic_/Interface/AddOns',
            Flavour.Classic,
        ),
        (
            '/foo/bar/_classic_ptr_/Interface/AddOns',
            Flavour.Classic,
        ),
        (
            '_classic_era_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            '_classic_era_ptr_/Interface/AddOns',
            Flavour.VanillaClassic,
        ),
        (
            'wowzerz/_retail_/Interface/AddOns',
            Flavour.Retail,
        ),
        (
            'anything goes',
            None,
        ),
    ],
)
def test_can_infer_flavour_from_addon_dir(
    path: str,
    flavour: Flavour | None,
):
    assert infer_flavour_from_addon_dir(path) is flavour


@pytest.mark.skipif(sys.platform != 'darwin', reason='Only supported on Mac')
def test_can_find_mac_installations(
    monkeypatch: pytest.MonkeyPatch,
):
    with monkeypatch.context() as patcher:

        def check_output_no_installation(*args, **kwargs):
            return ''

        patcher.setattr('subprocess.check_output', check_output_no_installation)

        assert not list(find_installations())

    with monkeypatch.context() as patcher:
        app_bundle_paths = {
            Path('/Applications/World of Warcraft/_retail_/World of Warcraft.app'): {
                'code': 'wow',
                'flavour': Flavour.Retail,
            },
            Path('/Applications/World of Warcraft/_classic_/World of Warcraft Classic.app'): {
                'code': 'wow_classic',
                'flavour': Flavour.Classic,
            },
        }

        def check_output_has_installations(*args, **kwargs):
            return '\n'.join(map(str, app_bundle_paths))

        patcher.setattr('subprocess.check_output', check_output_has_installations)

        assert list(find_installations()) == [(p.parent, f) for p, f in app_bundle_paths.items()]


def test_installation_version_extraction_from_addon_dir(
    iw_profile_config: ProfileConfig,
):
    iw_profile_config.addon_dir.parents[2].joinpath('.build.info').write_text("""\
Version!STRING:0|Product!STRING:0
10.9.8.7|wow
""")
    assert get_installation_version_from_addon_dir(iw_profile_config.addon_dir) == '10.9.8.7'



from __future__ import annotations

from pathlib import Path
from typing import Any

import aiohttp
import aiohttp.web
import pytest
from aresponses import ResponsesMockServer
from aresponses.errors import NoRouteFoundError
from yarl import URL

import instawow._logging
import instawow.config
import instawow.http
import instawow.shared_ctx
from instawow.shared_ctx import ConfigBoundCtx
from instawow.wow_installations import _DELECTABLE_DIR_NAMES, Flavour

from .fixtures.http import ROUTES


def pytest_addoption(parser: pytest.Parser):
    parser.addoption('--iw-no-mock-http', action='store_true')


@pytest.fixture(autouse=True)
def anyio_backend():
    return 'asyncio'


@pytest.fixture
def caplog(caplog: pytest.LogCaptureFixture):
    handler_id = instawow._logging.logger.add(
        caplog.handler,
        format='{message}',
        level=0,
        filter=lambda record: record['level'].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    instawow._logging.logger.remove(handler_id)


class _StrictResponsesMockServer(ResponsesMockServer):
    async def _find_response(self, request: aiohttp.web.Request):
        response = await super()._find_response(request)
        if response == (None, None):
            raise NoRouteFoundError(f'No match found for <{request.method} {request.url}>')
        return response


@pytest.fixture
async def iw_aresponses():
    async with _StrictResponsesMockServer() as server:
        yield server


@pytest.fixture(params=['foo'])
def iw_global_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    return {
        'config_dir': tmp_path / '__config__' / 'config',
        'temp_dir': tmp_path / '__config__' / 'temp',
        'state_dir': tmp_path / '__config__' / 'state',
        'access_tokens': {
            'cfcore': request.param,
            'github': None,
            'wago': None,
            'wago_addons': request.param,
        },
    }


@pytest.fixture(params=[Flavour.Retail])
def iw_profile_config_values(request: pytest.FixtureRequest, tmp_path: Path):
    installation_dir = (
        tmp_path
        / 'wow'
        / next(
            (k for k, v in _DELECTABLE_DIR_NAMES.items() if v['flavour'] is request.param),
            '_unknown_',
        )
    )
    addon_dir = installation_dir / 'interface' / 'addons'
    addon_dir.mkdir(parents=True)
    return {
        'profile': '__default__',
        'addon_dir': addon_dir,
        'game_flavour': request.param,
        '_installation_dir': installation_dir,
    }


@pytest.fixture
def iw_global_config(iw_global_config_values: dict[str, Any]):
    return instawow.config.GlobalConfig.from_values(iw_global_config_values).write()


@pytest.fixture
def iw_profile_config(
    iw_profile_config_values: dict[str, Any], iw_global_config: instawow.config.GlobalConfig
):
    return instawow.config.ProfileConfig.from_values(
        {'global_config': iw_global_config, **iw_profile_config_values}
    ).write()


@pytest.fixture(autouse=True)
def _iw_global_config_defaults(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
):
    for key in 'config_dir', 'temp_dir', 'state_dir':
        monkeypatch.setenv(f'INSTAWOW_{key.upper()}', str(iw_global_config_values[key]))


@pytest.fixture
async def iw_web_client(iw_global_config: instawow.config.GlobalConfig):
    async with instawow.http.init_web_client(iw_global_config.http_cache_dir) as web_client:
        yield web_client


@pytest.fixture
async def _iw_web_client_ctx(iw_web_client: instawow.http.ClientSession):
    instawow.shared_ctx.web_client_var.set(iw_web_client)


@pytest.fixture
def iw_config_ctx(iw_profile_config: instawow.config.ProfileConfig):
    with ConfigBoundCtx(iw_profile_config) as config_ctx:
        yield config_ctx


@pytest.fixture(autouse=True, params=['all'])
async def _iw_mock_aiohttp_requests(
    request: pytest.FixtureRequest, iw_aresponses: _StrictResponsesMockServer
):
    if request.config.getoption('--iw-no-mock-http') or any(
        m.name == 'iw_no_mock_http' for m in request.node.iter_markers()
    ):
        await iw_aresponses.__aexit__(*((None,) * 3))
        return

    if request.param == 'all':
        routes = ROUTES.values()
    else:
        urls = set(map(URL, request.param))
        if not urls.issubset(ROUTES.keys()):
            raise ValueError('Supplied routes must be subset of all routes')

        routes = (ROUTES[k] for k in ROUTES.keys() & urls)

    for route in routes:
        iw_aresponses.add(**route.to_aresponses_add_args())



from __future__ import annotations

import click

from instawow import cli


@click.group('weakauras-companion')
def wa_updater_command_group() -> None:
    "Manage your WeakAuras."


@wa_updater_command_group.command('build')
@click.pass_obj
def build_weakauras_companion(config_ctx: cli.ConfigBoundCtxProxy) -> None:
    "Build the WeakAuras Companion add-on."

    from ._core import WaCompanionBuilder

    cli.run_with_progress(WaCompanionBuilder(config_ctx.config).build())


@wa_updater_command_group.command('list')
@click.pass_obj
def list_installed_wago_auras(config_ctx: cli.ConfigBoundCtxProxy) -> None:
    "List WeakAuras installed from Wago."
    from instawow._utils.text import tabulate

    from ._core import WaCompanionBuilder

    builder = WaCompanionBuilder(config_ctx.config)
    builder.config.ensure_dirs()

    installed_auras = sorted(
        (g.addon_name, a.id, a.url)
        for g in builder.extract_installed_auras()
        for v in g.root.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('type', 'name', 'URL'), *installed_auras]))



from __future__ import annotations

import hashlib
import os
import re
from collections.abc import Iterator
from functools import lru_cache
from pathlib import Path

from .._logging import logger

_TOC_FILE_PATH_PATTERN = re.compile(
    r'^(?P<name>[^/]+)/(?P=name)(?:[-_](?:mainline|bcc|tbc|classic|vanilla|wrath|wotlkc))?\.toc$',
    flags=re.IGNORECASE,
)
_BINDINGS_XML_FILE_PATH_PATTERN = re.compile(
    r'^[^/]+/Bindings\.xml$',
    flags=re.IGNORECASE,
)
_TOC_COMMENT_PATTERN = re.compile(
    r'\s*#.*$',
    flags=re.IGNORECASE | re.MULTILINE,
)
_XML_COMMENT_PATTERN = re.compile(
    r'<!--.*?-->',
    flags=re.IGNORECASE | re.DOTALL,
)
_TOC_INCLUDE_PATTERN = re.compile(
    r'^\s*(?P<path>(?:(?<!\.\.).)+\.(?:xml|lua))\s*$',
    flags=re.IGNORECASE | re.MULTILINE,
)
_XML_INCLUDE_PATTERN = re.compile(
    r"<(?:Include|Script)\s+file=[\"'](?P<path>(?:(?<!\.\.).)+)[\"']\s*\/>",
    flags=re.IGNORECASE,
)

_INCLUDE_FILE_PATTERNS = {
    '.toc': (_TOC_COMMENT_PATTERN, _TOC_INCLUDE_PATTERN),
    '.xml': (_XML_COMMENT_PATTERN, _XML_INCLUDE_PATTERN),
}


def _scan_addon_folder(folder: Path, root_folder: Path) -> Iterator[Path]:
    for entry in map(Path, os.scandir(folder)):
        relative_posix_path = entry.relative_to(root_folder).as_posix()
        if _TOC_FILE_PATH_PATTERN.match(relative_posix_path):
            yield from _scan_includes(entry)
        elif _BINDINGS_XML_FILE_PATH_PATTERN.match(relative_posix_path):
            yield entry
        elif entry.is_dir():
            yield from _scan_addon_folder(entry, root_folder)


def _scan_includes(file_path: Path) -> Iterator[Path]:
    if not file_path.exists():
        logger.debug(f'{file_path} does not exist')
        return

    yield file_path

    lower_suffix = file_path.suffix.casefold()
    try:
        comment_pattern, include_pattern = _INCLUDE_FILE_PATTERNS[lower_suffix]
    except KeyError:
        return

    contents = file_path.read_text(encoding='utf-8-sig', errors='replace')
    contents_without_comments = comment_pattern.sub('', contents)
    matches = (
        m.group('path').replace('\\', os.sep).strip()
        for m in include_pattern.finditer(contents_without_comments)
    )
    for match in matches:
        yield from _scan_includes(file_path.parent / match)


def _md5hash_bytes(value: bytes):
    return hashlib.md5(value, usedforsecurity=False).hexdigest()


def _md5hash_file(path: Path):
    return _md5hash_bytes(path.read_bytes())


def _md5hash_hashes(value: Iterator[str]):
    return _md5hash_bytes(''.join(sorted(value)).encode())


@lru_cache
def generate_wowup_addon_hash(path: Path):
    files_to_hash = frozenset(_scan_addon_folder(path, path.parent))
    combined_hash = _md5hash_hashes(_md5hash_file(f) for f in files_to_hash)
    logger.debug(f'calculated {combined_hash} for {path.name}')
    return combined_hash



from __future__ import annotations

from itertools import product

import pytest

from instawow.pkg_archives import find_archive_addon_tocs, make_archive_member_filter_fn


def test_find_archive_addon_tocs_can_find_explicit_dirs():
    assert {h for _, h in find_archive_addon_tocs(['b/', 'b/b.toc'])} == {'b'}


def test_find_archive_addon_tocs_can_find_implicit_dirs():
    assert {h for _, h in find_archive_addon_tocs(['b/b.toc'])} == {'b'}


def test_find_archive_addon_tocs_discards_tocless_paths():
    assert {h for _, h in find_archive_addon_tocs(['a', 'b/b.toc', 'c/'])} == {'b'}


def test_find_archive_addon_tocs_discards_mismatched_tocs():
    assert not {h for _, h in find_archive_addon_tocs(['a', 'a/b.toc'])}


def test_find_archive_addon_tocs_accepts_multitoc():
    assert {h for _, h in find_archive_addon_tocs(['a', 'a/a_mainline.toc'])} == {'a'}


@pytest.mark.parametrize('ext', product('Tt', 'Oo', 'Cc'))
def test_find_archive_addon_tocs_toc_is_case_insensitive(ext: tuple[str, ...]):
    assert {h for _, h in find_archive_addon_tocs([f'a/a.{"".join(ext)}'])} == {'a'}


def test_make_archive_member_filter_fn_discards_names_with_prefix_not_in_dirs():
    is_member = make_archive_member_filter_fn({'b'})
    assert list(filter(is_member, ['a/', 'b/', 'aa/', 'bb/', 'b/c', 'a/d'])) == ['b/', 'b/c']



from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import pytest
from cattrs import AttributeValidationNote, ClassValidationError

from instawow.config import GlobalConfig, PluginConfig, ProfileConfig
from instawow.wow_installations import Flavour


def test_env_vars_always_take_precedence(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_AUTO_UPDATE_CHECK', '1')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    global_config = GlobalConfig.from_values(
        {**iw_global_config_values, 'auto_update_check': False}, env=True
    )
    profile_config = ProfileConfig.from_values(
        {'global_config': global_config, **iw_profile_config_values, 'game_flavour': 'retail'},
        env=True,
    )
    assert global_config.auto_update_check is True
    assert profile_config.game_flavour is Flavour.Classic


def test_read_profile_from_nonexistent_config_dir_raises(
    iw_global_config_values: dict[str, Any],
):
    global_config = GlobalConfig(config_dir=iw_global_config_values['config_dir'])
    with pytest.raises(FileNotFoundError):
        ProfileConfig.read(global_config, '__default__')


def test_init_with_nonexistent_addon_dir_raises(
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()

    values = dict(iw_profile_config_values)
    del values['_installation_dir']

    with pytest.raises(ValueError, match='not a writable directory'):
        ProfileConfig(**{'global_config': global_config, **values, 'addon_dir': '#@$foo'})


def test_default_config_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('XDG_CONFIG_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = GlobalConfig().config_dir
        assert config_dir == Path('/foo/instawow').resolve()

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')

        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'win32')

        patcher.delenv('APPDATA', False)
        assert GlobalConfig().config_dir == Path.home() / '.config' / 'instawow'

        patcher.setenv('APPDATA', '/foo')
        assert GlobalConfig().config_dir == Path('/foo/instawow').resolve()


def test_config_dir_xdg_env_var_is_respected_on_all_plats(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    config_parent_dir = tmp_path / 'instawow_config_parent_dir'
    monkeypatch.setenv('XDG_CONFIG_HOME', str(config_parent_dir))
    global_config = GlobalConfig()
    assert global_config.config_dir == config_parent_dir / 'instawow'


def test_config_dir_instawow_specific_env_var_takes_precedence(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    test_config_dir_xdg_env_var_is_respected_on_all_plats(monkeypatch, tmp_path)

    config_dir = tmp_path / 'instawow_config_dir'
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(config_dir))
    global_config = GlobalConfig.from_values(env=True)
    assert global_config.config_dir == config_dir


def test_default_state_dir(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv('XDG_STATE_HOME', False)

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')

        assert GlobalConfig().state_dir == Path.home() / '.local' / 'state' / 'instawow'

    for platform in {'darwin', 'win32'}:
        with monkeypatch.context() as patcher:
            patcher.setattr(sys, 'platform', platform)

            global_config = GlobalConfig()
            assert global_config.config_dir == global_config.state_dir


def test_state_dir_xdg_env_var_is_respected_on_all_plats(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    state_parent_dir = tmp_path / 'instawow_state_parent_dir'
    monkeypatch.setenv('XDG_STATE_HOME', str(state_parent_dir))
    global_config = GlobalConfig()
    assert global_config.state_dir == state_parent_dir / 'instawow'


def test_state_dir_instawow_specific_env_var_takes_precedence(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
):
    test_state_dir_xdg_env_var_is_respected_on_all_plats(monkeypatch, tmp_path)

    state_dir = tmp_path / 'instawow_state_dir'
    monkeypatch.setenv('INSTAWOW_STATE_DIR', str(state_dir))
    global_config = GlobalConfig.from_values(env=True)
    assert global_config.state_dir == state_dir


def test_can_list_profiles(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.read()
    assert set(global_config.iter_profiles()) == set()

    ProfileConfig.from_values({'global_config': global_config, **iw_profile_config_values}).write()
    ProfileConfig.from_values(
        {'global_config': global_config, **iw_profile_config_values, 'profile': 'foo'}, env=False
    ).write()
    assert set(global_config.iter_profiles()) == {'__default__', 'foo'}


def test_can_list_installations(
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.read()
    assert set(global_config.iter_installations()) == set()

    ProfileConfig.from_values({'global_config': global_config, **iw_profile_config_values}).write()
    assert set(global_config.iter_installations()) == {
        iw_profile_config_values['_installation_dir']
    }


def test_profile_dirs_are_populated(
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)
    profile_config = ProfileConfig.from_values(
        {'global_config': global_config, **iw_profile_config_values}
    ).write()
    assert {i.name for i in profile_config.config_dir.iterdir()} <= {'config.json'}
    assert {i.name for i in profile_config.state_dir.iterdir()} <= {'logs', 'plugins'}


def test_can_delete_profile(
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values).write()
    profile_config = ProfileConfig.from_values(
        {'global_config': global_config, **iw_profile_config_values}
    ).write()
    assert profile_config.config_dir.exists()
    profile_config.delete()
    assert not profile_config.config_dir.exists()


def test_validate_profile_name(
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)

    with pytest.raises(ClassValidationError) as exc_info:
        ProfileConfig.from_values(
            {'global_config': global_config, **iw_profile_config_values, 'profile': ''}
        )

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == ('Value must have a minimum length of 1',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class ProfileConfig @ attribute profile'
    assert type(note) is AttributeValidationNote
    assert note.name == 'profile'


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='chmod has no effect on Windows',
)
def test_validate_addon_dir(
    tmp_path: Path,
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)

    non_writeable_dir = tmp_path / 'non-writeable-dir'
    non_writeable_dir.mkdir(0o400)

    with pytest.raises(ClassValidationError) as exc_info:
        ProfileConfig.from_values(
            {
                'global_config': global_config,
                **iw_profile_config_values,
                'addon_dir': non_writeable_dir,
            }
        )

    (value_error,) = exc_info.value.exceptions
    assert value_error.args == (f'"{non_writeable_dir}" is not a writable directory',)

    (note,) = value_error.__notes__
    assert note == 'Structuring class ProfileConfig @ attribute addon_dir'
    assert type(note) is AttributeValidationNote
    assert note.name == 'addon_dir'


def test_plugin_cache_dirs(
    iw_global_config_values: dict[str, Any],
    iw_profile_config_values: dict[str, Any],
):
    global_config = GlobalConfig.from_values(iw_global_config_values)
    profile_config = ProfileConfig.from_values(
        {'global_config': global_config, **iw_profile_config_values}
    )
    plugin_config = PluginConfig(profile_config=profile_config, plugin='test')
    assert plugin_config.cache_dir == global_config.plugins_cache_dir / 'test'
    assert (
        plugin_config.profile_cache_dir
        == global_config.plugins_cache_dir / 'test' / 'profiles' / profile_config.profile
    )



from __future__ import annotations

from pathlib import Path

from instawow._utils.compat import fauxfrozen
from instawow.config import PluginConfig as _PluginConfig


@fauxfrozen
class PluginConfig(_PluginConfig):
    @property
    def access_token(self) -> str | None:
        return self.profile_config.global_config.access_tokens.wago

    @property
    def addon_zip_file(self) -> Path:
        return self.profile_cache_dir / 'WeakAurasCompanion.zip'

    @property
    def changelog_file(self) -> Path:
        return self.profile_cache_dir / 'CHANGELOG.md'

    @property
    def version_file(self) -> Path:
        return self.profile_cache_dir / 'version.txt'



from __future__ import annotations

import os
import sys
from collections.abc import Iterator
from enum import Enum
from functools import cache
from pathlib import Path
from typing import TypedDict, TypeVar

from typing_extensions import NotRequired, Self

from ._utils.compat import StrEnum
from ._utils.iteration import fill

_TEnum = TypeVar('_TEnum', bound=Enum)


class _FlavourMeta(TypedDict):
    aliases: NotRequired[list[str]]
    is_retired: NotRequired[bool]


_DummyFlavourMeta = _FlavourMeta()


class Flavour(StrEnum):
    # The current Classic version is always aliased to "classic".
    # The assumption here is that should Classic not be discontinued
    # it will continue to be updated in place so that new Classic versions
    # will inherit the "_classic_" folder.  This means we won't have to
    # migrate Classic profiles either automatically or by requiring user
    # intervention for new Classic releases.
    Retail = ('retail',)
    VanillaClassic = ('vanilla_classic',)
    Classic = ('classic', _FlavourMeta(aliases=['cataclysm_classic']))
    WrathClassic = ('wrath_classic', _FlavourMeta(is_retired=True))

    is_retired: bool

    def __new__(cls, value: str, flavour_meta: _FlavourMeta = _DummyFlavourMeta) -> Self:
        self = str.__new__(cls, value)
        self._value_ = value
        for alias in flavour_meta.get('aliases', []):
            cls._value2member_map_.setdefault(alias, self)
        self.is_retired = flavour_meta.get('is_retired', False)
        return self

    @classmethod
    def from_flavour_keyed_enum(cls, flavour_keyed_enum: Enum) -> Self:
        return cls[flavour_keyed_enum.name]

    def to_flavour_keyed_enum(self, flavour_keyed_enum: type[_TEnum]) -> _TEnum:
        return flavour_keyed_enum[self.name]

    def get_flavour_groups(self, affine: bool) -> list[tuple[Flavour, ...] | None]:
        match (self, affine):
            case (self.__class__.Classic, True):
                return [(self, self.__class__.WrathClassic), None]
            case (_, True):
                return [(self,), None]
            case _:
                return [(self,)]


class FlavourVersionRange(Enum):
    Retail = (
        range(1_00_00, 1_13_00),
        range(2_00_00, 2_05_00),
        range(3_00_00, 3_04_00),
        range(4_00_00, 4_04_00),
        range(5_00_00, 12_00_00),
    )
    VanillaClassic = (range(1_13_00, 2_00_00),)
    Classic = (range(4_04_00, 5_00_00),)
    WrathClassic = (range(3_04_00, 4_00_00),)

    @classmethod
    def _parse_version_string(cls, version_string: str) -> int:
        major, minor, patch = fill(map(int, version_string.split('.')), 0, 3)
        return major * 1_00_00 + minor * 1_00 + patch

    @classmethod
    def from_version(cls, version: int | str) -> Self | None:
        version_number = (
            version if isinstance(version, int) else cls._parse_version_string(version)
        )
        return next((f for f in cls if f.contains(version_number)), None)

    def contains(self, version: int | str) -> bool:
        version_number = (
            version if isinstance(version, int) else self._parse_version_string(version)
        )
        return any(version_number in r for r in self.value)


class _Product(TypedDict):
    code: str
    flavour: Flavour


_DELECTABLE_DIR_NAMES: dict[str, _Product] = {
    '_retail_': {
        'code': 'wow',
        'flavour': Flavour.Retail,
    },
    '_ptr_': {
        'code': 'wowt',
        'flavour': Flavour.Retail,
    },
    '_xptr_': {
        'code': 'wowxptr',
        'flavour': Flavour.Retail,
    },
    '_beta_': {
        'code': 'wow_beta',
        'flavour': Flavour.Retail,
    },
    '_classic_era_': {
        'code': 'wow_classic_era',
        'flavour': Flavour.VanillaClassic,
    },
    '_classic_era_ptr_': {
        'code': 'wow_classic_era_ptr',
        'flavour': Flavour.VanillaClassic,
    },
    # '_classic_era_beta_': {
    #     'code': 'wow_classic_era_beta',
    #     'flavour': Flavour.VanillaClassic,
    # },
    '_classic_': {
        'code': 'wow_classic',
        'flavour': Flavour.Classic,
    },
    '_classic_ptr_': {
        'code': 'wow_classic_ptr',
        'flavour': Flavour.Classic,
    },
    '_classic_beta_': {
        'code': 'wow_classic_beta',
        'flavour': Flavour.Classic,
    },
}

ADDON_DIR_PARTS = ('Interface', 'AddOns')
_NORMALISED_ADDON_DIR_PARTS = tuple(map(str.casefold, ADDON_DIR_PARTS))


def _find_mac_installations():
    import subprocess

    try:
        possible_installations = subprocess.check_output(
            [
                'mdfind',
                "kMDItemCFBundleIdentifier == 'com.blizzard.worldofwarcraft'",
            ],
            text=True,
            timeout=3,
        ).splitlines()
    except subprocess.TimeoutExpired:
        pass
    else:
        for match in possible_installations:
            installation_path = Path(match).parent
            yield (installation_path, _DELECTABLE_DIR_NAMES.get(installation_path.name))


def find_installations() -> Iterator[tuple[Path, _Product | None]]:
    if sys.platform == 'darwin':
        yield from _find_mac_installations()


def _read_info(info_path: Path):
    with open(info_path, encoding='utf-8') as info:
        header = next(info)
        fields = tuple(f[: f.find('!')] for f in header.rstrip().split('|'))
        return [dict(zip(fields, e.rstrip().split('|'))) for e in info if not e.startswith('#')]


@cache
def _get_installed_versions(outer_installation_path: Path):
    return {
        e['Product']: e['Version'] for e in _read_info(outer_installation_path / '.build.info')
    }


def get_installation_dir_from_addon_dir(path_like: os.PathLike[str] | str) -> Path | None:
    path = Path(path_like)
    tail = tuple(map(str.casefold, path.parts[-3:]))
    if len(tail) == 3 and tail[1:] == _NORMALISED_ADDON_DIR_PARTS:
        return path.parents[1]


def infer_flavour_from_addon_dir(path: os.PathLike[str] | str) -> Flavour | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        try:
            return _DELECTABLE_DIR_NAMES[maybe_installation_dir.name]['flavour']
        except KeyError:
            return None


def get_installation_version_from_addon_dir(path: os.PathLike[str] | str) -> str | None:
    maybe_installation_dir = get_installation_dir_from_addon_dir(path)
    if maybe_installation_dir:
        product = _DELECTABLE_DIR_NAMES.get(maybe_installation_dir.name)
        if product:
            all_versions = _get_installed_versions(maybe_installation_dir.parent)
            return all_versions[product['code']]



from __future__ import annotations

import asyncio
import importlib.resources
import sys
import time
from pathlib import Path

import pytest

from instawow._utils.aio import run_in_thread
from instawow._utils.iteration import bucketise, merge_intersecting_sets
from instawow._utils.text import tabulate
from instawow._utils.web import file_uri_to_path
from instawow.matchers.addon_toc import TocReader


@pytest.fixture
def fake_addon_toc():
    with importlib.resources.as_file(
        importlib.resources.files(__spec__.parent) / 'fixtures' / 'FakeAddon' / 'FakeAddon.toc'
    ) as file:
        yield file


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


def test_tabulate_spits_out_ascii_table(fake_addon_toc: Path):
    toc_reader = TocReader.from_path(fake_addon_toc)
    data = [('key', 'value'), *toc_reader.items()]
    assert tabulate(data) == (
        '  key        value    \n'
        '-------  -------------\n'
        'Normal   Normal entry \n'
        'Compact  Compact entry'
    )


def test_merge_intersecting_sets_in_noncontiguous_collection():
    collection = [
        {'a'},
        {'b', 'c'},
        {'a', 'd'},
        {'e'},
        {'c', 'f'},
        {'g', 'a'},
    ]
    output = [
        {'a', 'd', 'g'},
        {'b', 'c', 'f'},
        {'e'},
    ]
    assert sorted(merge_intersecting_sets(collection)) == output


@pytest.mark.skipif(sys.platform == 'win32', reason='platform dependent')
def test_file_uri_to_path_posix_leading_slash_is_preserved():
    uri = Path('/foo/bar').as_uri()
    assert uri == 'file:///foo/bar'
    assert file_uri_to_path(uri) == '/foo/bar'


@pytest.mark.skipif(sys.platform != 'win32', reason='platform dependent')
def test_file_uri_to_path_win32_leading_slash_is_stripped():
    uri = Path('C:/foo/bar').as_uri()
    assert uri == 'file:///C:/foo/bar'
    assert file_uri_to_path(uri) == 'C:/foo/bar'


async def test_generator_in_run_in_thread_does_not_lock_up_loop():
    def foo():
        time.sleep(2)
        yield 'foo'

    async def bar():
        await asyncio.sleep(1)
        return ['bar']

    assert [await a for a in asyncio.as_completed([run_in_thread(list)(foo()), bar()])] == [
        ['bar'],
        ['foo'],
    ]



from __future__ import annotations

import instawow.plugins

from ._cli import wa_updater_command_group


@instawow.plugins.hookimpl
def instawow_add_commands():
    return (wa_updater_command_group,)



from __future__ import annotations

import contextvars as cv
from collections.abc import Mapping
from contextlib import AbstractAsyncContextManager, AbstractContextManager, ExitStack
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, TypeAlias

from . import _sources, http, pkg_db
from ._utils.iteration import WeakValueDefaultDictionary
from .config import ProfileConfig
from .plugins import get_plugin_resolvers
from .resolvers import Resolvers


class _DummyLock(AbstractAsyncContextManager[None]):
    async def __aexit__(self, *args: object):
        return None


Locks: TypeAlias = Mapping[object, AbstractAsyncContextManager[None]]

locks_var = cv.ContextVar[Locks]('locks_var', default=WeakValueDefaultDictionary(_DummyLock))
web_client_var = cv.ContextVar['http.ClientSession']('web_client_var')


def _database_from_config(config: ProfileConfig):
    return pkg_db.DatabaseHandle(config.db_file)


def _resolvers_from_config(config: ProfileConfig):
    builtin_resolver_classes = list(_sources.DEFAULT_RESOLVERS)

    for resolver, access_token in (
        (r, getattr(config.global_config.access_tokens, r.requires_access_token, None))
        for r in _sources.DEFAULT_RESOLVERS
        if r.requires_access_token is not None
    ):
        if access_token is None:
            builtin_resolver_classes.remove(resolver)

    resolver_classes = chain(
        (r for g in get_plugin_resolvers() for r in g), builtin_resolver_classes
    )
    return Resolvers((r.metadata.id, r(config)) for r in resolver_classes)


class ConfigBoundCtx(AbstractContextManager['ConfigBoundCtx']):
    def __init__(self, config: ProfileConfig) -> None:
        self._exit_stack = ExitStack()

        self.config: ProfileConfig = config

    def __exit__(self, *args: object) -> None:
        self._exit_stack.close()

    @cached_property
    def database(self) -> pkg_db.DatabaseHandle:
        database = _database_from_config(self.config)
        self._exit_stack.callback(database.close)
        return database

    @cached_property
    def resolvers(self) -> Resolvers:
        return _resolvers_from_config(self.config)


if TYPE_CHECKING:
    locks: Locks
    web_client: http.ClientSession
else:

    def __getattr__(name: str):
        match name:
            case 'locks':
                return locks_var.get()
            case 'web_client':
                return web_client_var.get()
            case _:
                raise AttributeError(name)



from __future__ import annotations

from datetime import datetime, timezone
from typing import final

import click

import instawow.plugins
from instawow.definitions import ChangelogFormat, Defn, SourceMetadata
from instawow.resolvers import BaseResolver, PkgCandidate


@click.command
def foo():
    "don't foo where you bar"
    print('success!')


@final
class MyResolver(BaseResolver):
    metadata = SourceMetadata(
        id='me',
        name="It's me",
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Markdown,
        addon_toc_key=None,
    )
    requires_access_token = None

    async def _resolve_one(self, defn: Defn, metadata: None) -> PkgCandidate:
        return PkgCandidate(
            id='bar',
            slug='bar',
            name='Bar',
            description='The quintessential bar add-on, brought to you by yours truly',
            url='http://example.com/',
            download_url='file:///...',
            date_published=datetime.now(timezone.utc),
            version='0',
            changelog_url='data:,',
        )


@instawow.plugins.hookimpl
def instawow_add_commands():
    return (foo,)


@instawow.plugins.hookimpl
def instawow_add_resolvers():
    return (MyResolver,)



from __future__ import annotations

import sys
from functools import partial
from typing import TYPE_CHECKING, Any

import attrs

if sys.version_info >= (3, 11):
    from enum import StrEnum as StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        @staticmethod
        def _generate_next_value_(name: str, start: int, count: int, last_values: list[object]):
            return name.lower()


def add_exc_note(exc: BaseException, note: str) -> None:
    if sys.version_info >= (3, 11):
        exc.add_note(note)

    else:
        exc_without_notes: Any = exc

        if not hasattr(exc, '__notes__'):
            exc_without_notes.__notes__ = list[str]()
        exc_without_notes.__notes__.append(note)


if TYPE_CHECKING:
    fauxfrozen = attrs.frozen
else:
    fauxfrozen = partial(attrs.define, unsafe_hash=True)



from __future__ import annotations

import asyncio
import concurrent.futures
import contextlib
import contextvars
import os
from collections.abc import Callable, Coroutine, Set
from functools import wraps
from typing import Any, TypeVar, cast

import aiohttp_client_cache
import diskcache
from typing_extensions import ParamSpec

_U = TypeVar('_U')
_P = ParamSpec('_P')


@contextlib.asynccontextmanager
async def make_cache(cache_dir: os.PathLike[str]):
    with concurrent.futures.ThreadPoolExecutor(1, '_http_cache') as executor:
        loop = asyncio.get_running_loop()

        def run_in_thread2(fn: Callable[_P, _U]) -> Callable[_P, Coroutine[Any, Any, _U]]:
            @wraps(fn)
            async def wrapper(*args: _P.args, **kwargs: _P.kwargs):
                return await loop.run_in_executor(
                    executor, lambda: contextvars.copy_context().run(fn, *args, **kwargs)
                )

            return wrapper

        # diskcache will only close the sqlite connection if it was initialised
        # in the same thread.
        cache = await run_in_thread2(diskcache.Cache)(os.fspath(cache_dir))

        class Cache(aiohttp_client_cache.BaseCache):
            @run_in_thread2
            def bulk_delete(self, keys: Set[str]):
                for key in keys:
                    cache.delete(key)

            @run_in_thread2
            def clear(self):
                cache.clear()

            @run_in_thread2
            def contains(self, key: str):
                return key in cache

            @run_in_thread2
            def delete(self, key: str):
                cache.delete(key)

            async def keys(self):
                with contextlib.suppress(StopIteration):
                    iter_keys = iter(cache)

                    while True:
                        yield (await run_in_thread2(next)(iter_keys))

            @run_in_thread2
            def read(self, key: str):
                return cache.get(key)

            @run_in_thread2
            def size(self):
                return len(cache)

            async def values(self):
                async for key in self.keys():
                    yield cast(aiohttp_client_cache.ResponseOrKey, run_in_thread2(cache.get)(key))

            @run_in_thread2
            def write(self, key: str, item: aiohttp_client_cache.ResponseOrKey):
                cache[key] = item

        try:
            yield Cache()
        finally:
            await run_in_thread2(cache.close)()



from __future__ import annotations

import importlib.resources
import json
from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import timedelta
from functools import reduce
from itertools import chain, product
from typing import Literal, TypeAlias, cast

import cattrs
from typing_extensions import NotRequired as N
from typing_extensions import TypedDict
from yarl import URL

from instawow import http, shared_ctx
from instawow._logging import logger
from instawow._progress_reporting import make_default_progress, make_incrementing_progress_tracker
from instawow._utils.aio import gather, run_in_thread
from instawow._utils.compat import StrEnum, fauxfrozen
from instawow._utils.iteration import bucketise
from instawow._utils.perf import time_op
from instawow._utils.text import shasum
from instawow.config import ProfileConfig
from instawow.wow_installations import get_installation_dir_from_addon_dir

from ._config import PluginConfig

_LuaTable: TypeAlias = Mapping[str, '_LuaTable']
_Auras: TypeAlias = 'WeakAuras | Plateroos'
_Match: TypeAlias = tuple[Sequence['WeakAura | Plateroo'], '_WagoApiResponse', str]


class _WagoApiResponse(TypedDict):
    _id: str  # +   # Alphanumeric ID
    name: str  # +  # User-facing name
    slug: str  # +  # Slug if it has one; otherwise same as ``_id``
    url: str
    created: str  # ISO datetime
    modified: str  # ISO datetime
    game: str  # "classic" or xpac, e.g. "bfa"
    username: N[str]  # +  # Author username
    version: int  # +   # Version counter, incremented with every update
    # Semver auto-generated from ``version`` - for presentation only
    versionString: str
    changelog: _WagoApiResponse_Changelog  # +
    forkOf: N[str]  # Only present on forks
    regionType: N[str]  # Only present on WAs


class _WagoApiResponse_Changelog(TypedDict):
    format: N[Literal['bbcode', 'markdown']]
    text: N[str]


_CHECK_API_URL = URL('https://data.wago.io/api/check')
_IMPORT_STRING_API_URL = URL('https://data.wago.io/api/raw/encoded')

_aura_converter = cattrs.Converter()
_aura_converter.register_structure_hook(URL, lambda v, _: URL(v))
_aura_converter.register_unstructure_hook(URL, str)


@fauxfrozen(kw_only=True)
class WeakAura:
    id: str
    uid: str
    parent: str | None = None
    url: URL
    version: int

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        url_string = lua_table.get('url')
        if url_string is not None:
            weakaura = _aura_converter.structure(lua_table, cls)
            if weakaura.url.host == 'wago.io':
                return weakaura


@fauxfrozen(kw_only=True)
class Plateroo(WeakAura):
    uid: str = ''

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        return super().from_lua_table({**lua_table, 'id': lua_table['Name']})


@fauxfrozen
class WeakAuras:
    api_ep = 'weakauras'
    addon_name = 'WeakAuras'

    root: dict[str, list[WeakAura]]

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        raw_auras = lua_table['WeakAurasSaved']['displays']
        auras = (WeakAura.from_lua_table(t) for t in raw_auras.values())
        sorted_auras = sorted(filter(None, auras), key=lambda a: a.id)
        return cls(bucketise(sorted_auras, key=lambda a: a.url.parts[1]))


@fauxfrozen
class Plateroos:
    api_ep = 'plater'
    addon_name = 'Plater'

    root: dict[str, list[Plateroo]]

    @classmethod
    def from_lua_table(cls, lua_table: _LuaTable):
        raw_auras = lua_table['PlaterDB']['profiles']
        auras = (
            Plateroo.from_lua_table(cast(_LuaTable, t))
            for n, p in raw_auras.items()
            for t in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                p.get('script_data') or (),
                p.get('hook_data') or (),
            )
        )
        sorted_auras = sorted(filter(None, auras), key=lambda a: a.id)
        return cls({a.url.parts[1]: [a] for a in sorted_auras})


def _merge_auras(auras: Iterable[_Auras]):
    return {
        t: t(reduce(lambda a, b: a | b, (i.root for i in a)))
        for t, a in bucketise(auras, key=type).items()
    }


class _TocNumber(StrEnum):
    Retail = '100206'
    VanillaClassic = '11502'
    Classic = '40400'
    WrathClassic = '30403'


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, profile_config: ProfileConfig) -> None:
        self._profile_config = profile_config
        self.config = PluginConfig(self._profile_config, __spec__.parent)

    def _make_request_headers(self):
        access_token = self.config.access_token
        if access_token:
            return {'api-key': access_token}

    @staticmethod
    def extract_auras(model: type[_Auras], source: str) -> _Auras:
        from ._custom_slpp import loads

        return model.from_lua_table(loads(f'{{{source}}}'))

    def extract_installed_auras(self) -> Iterator[_Auras]:
        installation_dir = get_installation_dir_from_addon_dir(self._profile_config.addon_dir)
        if not installation_dir:
            raise ValueError(
                f'cannot extract installation folder from {self._profile_config.addon_dir}'
            )

        saved_vars_by_account = installation_dir.glob('WTF/Account/*/SavedVariables')
        for saved_vars, model in product(
            saved_vars_by_account,
            [WeakAuras, Plateroos],
        ):
            file = (saved_vars / model.addon_name).with_suffix('.lua')
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                content = file.read_text(encoding='utf-8-sig', errors='replace')
                aura_group_cache = self.config.cache_dir / shasum(content)
                if aura_group_cache.exists():
                    logger.info(f'loading {file} from cache at {aura_group_cache}')
                    aura_group_json = json.loads(aura_group_cache.read_bytes())
                    aura_group = _aura_converter.structure({'root': aura_group_json}, model)
                else:
                    with time_op(
                        lambda t: logger.debug(
                            f'extracted {model.__name__} in {t:.3f}s'  # noqa: B023
                        )
                    ):
                        aura_group = self.extract_auras(model, content)
                    aura_group_cache.write_text(
                        json.dumps(_aura_converter.unstructure(aura_group.root)),
                        encoding='utf-8',
                    )
                yield aura_group

    async def _fetch_wago_metadata(self, api_ep: str, aura_ids: Iterable[str]):
        async with shared_ctx.web_client.post(
            (_CHECK_API_URL / api_ep),
            expire_after=timedelta(minutes=30),
            headers=self._make_request_headers(),
            json={'ids': list(aura_ids)},
            trace_request_ctx={
                'progress': make_default_progress(type_='download', label='Fetching aura metadata')
            },
        ) as response:
            metadata: list[_WagoApiResponse]
            if response.status == 404:
                metadata = []
                return metadata

            response.raise_for_status()

            metadata = await response.json()
            return sorted(metadata, key=lambda r: r['slug'])

    async def _fetch_wago_import_string(self, remote_aura: _WagoApiResponse):
        async with shared_ctx.web_client.get(
            _IMPORT_STRING_API_URL.with_query(id=remote_aura['_id']).with_fragment(
                str(remote_aura['version'])
            ),
            expire_after=http.CACHE_INDEFINITELY,
            headers=self._make_request_headers(),
            raise_for_status=True,
        ) as response:
            return await response.text()

    async def get_remote_auras(self, aura_group: _Auras) -> list[_Match]:
        if not aura_group.root:
            return []

        metadata = await self._fetch_wago_metadata(aura_group.api_ep, aura_group.root)

        track_progress = make_incrementing_progress_tracker(
            len(metadata), f'Fetching import strings: {aura_group.addon_name}'
        )

        import_strings = await gather(
            track_progress(self._fetch_wago_import_string(r)) for r in metadata
        )
        return [
            (aura_group.root.get(r['slug']) or aura_group.root[r['_id']], r, i)
            for r, i in zip(metadata, import_strings)
        ]

    def _generate_addon(self, auras: Iterable[tuple[type[_Auras], list[_Match]]]):
        from zipfile import ZipFile, ZipInfo

        from . import _templates

        template_resources = importlib.resources.files(_templates)

        aura_dict = dict.fromkeys((WeakAuras, Plateroos), list[_Match]()) | dict(auras)

        with ZipFile(self.config.addon_zip_file, 'w') as file:

            def write_file(filename: str, content: str):
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp would be set to the current time
                # which would render the build unreproducible.
                file.writestr(
                    ZipInfo(filename=f'WeakAurasCompanion/{filename}'),
                    content,
                )
                return content

            def make_slug_entry(metadata: _WagoApiResponse, import_string: str):
                return f"""\
            [ [=[{metadata['slug']}]=] ] = {{
                name = [=[{metadata['name']}]=],
                author = [=[{metadata.get('username', '__unknown__')}]=],
                encoded = [=[{import_string}]=],
                wagoVersion = [=[{metadata['version']}]=],
                wagoSemver = [=[{metadata['version']}]=],
                versionNote = [=[{metadata['changelog'].get('text', '')}]=],
                source = [=[Wago]=],
            }},"""

            NL = '\n'
            data_output = write_file(
                'data.lua',
                f'''\
-- file generated automatically
WeakAurasCompanionData = {{
{NL.join(f"""
    {c.addon_name} = {{
        slugs = {{
{NL.join(make_slug_entry(m, i) for _, m, i in v)}
        }},
        stash = {{
        }},
        stopmotionFiles = {{
        }},
    }},"""
    for c, v in aura_dict.items()
)}
}}
''',
            )

            init_output = write_file(
                'init.lua', template_resources.joinpath('init.lua').read_text()
            )

            interface_version = self._profile_config.game_flavour.to_flavour_keyed_enum(_TocNumber)
            addon_version = shasum(data_output, init_output, interface_version)[:7]

            toc_tpl = template_resources.joinpath('WeakAurasCompanion.toc').read_text()
            write_file(
                'WeakAurasCompanion.toc',
                toc_tpl.format(
                    interface=self._profile_config.game_flavour.to_flavour_keyed_enum(_TocNumber),
                    version=addon_version,
                ),
            )

        changelog_tpl = template_resources.joinpath('CHANGELOG.md').read_text()
        self.config.changelog_file.write_text(
            '\n\n'.join(
                changelog_tpl.format(
                    name=a.id,
                    url=a.url.parent,
                    version=metadata['version'],
                    changelog=metadata['changelog'].get('text') or 'n/a',
                )
                for v in aura_dict.values()
                for existing_auras, metadata, _ in v
                for a in (next((i for i in existing_auras if not i.parent), existing_auras[0]),)
                if a.version != metadata['version']
            )
            or 'n/a',
            encoding='utf-8',
        )

        self.config.version_file.write_text(
            addon_version,
            encoding='utf-8',
        )

    async def build(self) -> None:
        await run_in_thread(self.config.ensure_dirs)()

        installed_auras_by_type = _merge_auras(
            await run_in_thread(list[_Auras])(self.extract_installed_auras()),
        )
        remote_auras = await gather(
            self.get_remote_auras(r) for r in installed_auras_by_type.values()
        )
        await run_in_thread(self._generate_addon)(zip(installed_auras_by_type, remote_auras))

    def get_version(self) -> str:
        return self.config.version_file.read_text(encoding='utf-8')



from __future__ import annotations

import enum
import re
from collections.abc import Awaitable, Iterable, Mapping
from itertools import chain, product
from pathlib import Path
from typing import Protocol

import attrs
from typing_extensions import Self

from .. import shared_ctx
from .._utils.aio import gather
from .._utils.compat import fauxfrozen
from .._utils.iteration import bucketise, merge_intersecting_sets, uniq
from ..catalogue import synchronise as synchronise_catalogue
from ..definitions import Defn
from ..wow_installations import Flavour
from ._addon_hashing import generate_wowup_addon_hash
from .addon_toc import TocReader


class Matcher(Protocol):  # pragma: no cover
    def __call__(
        self, config_ctx: shared_ctx.ConfigBoundCtx, leftovers: frozenset[AddonFolder]
    ) -> Awaitable[list[tuple[list[AddonFolder], list[Defn]]]]: ...


class AddonHashMethod(enum.Enum):
    Wowup = enum.auto()


# https://github.com/Stanzilla/WoWUIBugs/issues/68#issuecomment-830351390
# https://warcraft.wiki.gg/wiki/TOC_format#Multiple_client_flavors
FLAVOUR_TOC_IDS = {
    Flavour.Retail: ('Mainline',),
    Flavour.VanillaClassic: ('Vanilla', 'Classic'),
    Flavour.Classic: ('Cata', 'Classic'),
    Flavour.WrathClassic: ('Wrath', 'WOTLKC', 'Classic'),
}

FLAVOUR_TOC_SUFFIXES = {
    k: tuple(f'{s}{f}.toc' for s, f in product('-_', v)) for k, v in FLAVOUR_TOC_IDS.items()
}
NORMALISED_FLAVOUR_TOC_SUFFIXES = {
    k: tuple(i.lower() for i in v) for k, v in FLAVOUR_TOC_SUFFIXES.items()
}


@fauxfrozen(order=True)
class AddonFolder:
    path: Path = attrs.field(eq=False, order=False)
    toc_reader: TocReader = attrs.field(eq=False, order=False)
    name: str = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        object.__setattr__(self, 'name', self.path.name)

    @classmethod
    def from_path(cls, flavour: Flavour, path: Path) -> Self | None:
        for suffix in chain(FLAVOUR_TOC_SUFFIXES[flavour], ('.toc',)):
            try:
                toc_reader = TocReader.from_path(path / (path.name + suffix))
                return cls(path, toc_reader)
            except FileNotFoundError:
                pass

    def get_defns_from_toc_keys(self, keys_and_ids: Iterable[tuple[str, str]]) -> frozenset[Defn]:
        return frozenset(
            Defn(s, i) for k, s in keys_and_ids for i in (self.toc_reader.get(k),) if i
        )


def hash_addon_contents(addon_path: Path, method: AddonHashMethod):
    match method:
        case AddonHashMethod.Wowup:
            return generate_wowup_addon_hash(addon_path)


def _get_unreconciled_folders(config_ctx: shared_ctx.ConfigBoundCtx):
    with config_ctx.database.connect() as connection:
        pkg_folders = [n for (n,) in connection.execute('SELECT name FROM pkg_folder').fetchall()]

    unreconciled_folder_paths = (
        p
        for p in config_ctx.config.addon_dir.iterdir()
        if p.name not in pkg_folders and p.is_dir() and not p.is_symlink()
    )
    for path in unreconciled_folder_paths:
        addon_folder = AddonFolder.from_path(config_ctx.config.game_flavour, path)
        if addon_folder:
            yield addon_folder


def get_unreconciled_folders(config_ctx: shared_ctx.ConfigBoundCtx) -> frozenset[AddonFolder]:
    return frozenset(_get_unreconciled_folders(config_ctx))


async def match_toc_source_ids(
    config_ctx: shared_ctx.ConfigBoundCtx, leftovers: frozenset[AddonFolder]
):
    catalogue = await synchronise_catalogue()

    def get_catalogue_defns(extracted_defns: Iterable[Defn]):
        for defn in extracted_defns:
            entry = catalogue.keyed_entries.get((defn.source, defn.alias))
            if entry:
                for addon_key in entry.same_as:
                    if addon_key.source in config_ctx.resolvers:
                        yield Defn(addon_key.source, addon_key.id)

    def get_addon_and_defn_pairs():
        for addon in sorted(leftovers):
            defns = addon.get_defns_from_toc_keys(config_ctx.resolvers.addon_toc_key_and_id_pairs)
            if defns:
                yield (addon, defns | frozenset(get_catalogue_defns(defns)))

    matches = list(get_addon_and_defn_pairs())
    merged_defns_by_constituent_defn = {
        i: s for s in merge_intersecting_sets(d for _, d in matches) for i in s
    }
    folders_grouped_by_overlapping_defns = bucketise(
        matches,
        lambda i: merged_defns_by_constituent_defn[next(iter(i[1]))],
    )
    return [
        (
            [n for n, _ in f],
            sorted(b, key=lambda d: config_ctx.resolvers.priority_dict[d.source]),
        )
        for b, f in folders_grouped_by_overlapping_defns.items()
    ]


async def match_folder_hashes(
    config_ctx: shared_ctx.ConfigBoundCtx, leftovers: frozenset[AddonFolder]
):
    matches = await gather(
        r.get_folder_hash_matches(leftovers) for r in config_ctx.resolvers.values()
    )
    flattened_matches = [t for g in matches for t in g]
    merged_folders_by_constituent_folder = {
        i: s for s in merge_intersecting_sets(f for _, f in flattened_matches) for i in s
    }
    matches_grouped_by_overlapping_folder_names = bucketise(
        flattened_matches, lambda v: merged_folders_by_constituent_folder[next(iter(v[1]))]
    )
    return sorted(
        (
            sorted(f),
            sorted(
                uniq(d for d, _ in b), key=lambda d: config_ctx.resolvers.priority_dict[d.source]
            ),
        )
        for f, b in matches_grouped_by_overlapping_folder_names.items()
    )


async def match_folder_name_subsets(
    config_ctx: shared_ctx.ConfigBoundCtx, leftovers: frozenset[AddonFolder]
):
    catalogue = await synchronise_catalogue()

    leftovers_by_name = {e.name: e for e in leftovers}

    matches = [
        (frozenset(leftovers_by_name[n] for n in m), Defn(i.source, i.id))
        for i in catalogue.entries
        if config_ctx.config.game_flavour in i.game_flavours
        for f in i.folders
        for m in (f & leftovers_by_name.keys(),)
        if m
    ]

    merged_folders_by_constituent_folder = {
        i: s for s in merge_intersecting_sets(f for f, _ in matches) for i in s
    }
    matches_grouped_by_overlapping_folder_names = bucketise(
        matches, lambda v: merged_folders_by_constituent_folder[next(iter(v[0]))]
    )

    def sort_key(folders: frozenset[AddonFolder], defn: Defn):
        return (-len(folders), config_ctx.resolvers.priority_dict[defn.source])

    return [
        (sorted(f), uniq(d for _, d in sorted(b, key=lambda v: sort_key(*v))))
        for f, b in matches_grouped_by_overlapping_folder_names.items()
    ]


async def match_addon_names_with_folder_names(
    config_ctx: shared_ctx.ConfigBoundCtx, leftovers: frozenset[AddonFolder]
):
    def normalise(value: str):
        return re.sub(r'[^0-9A-Za-z]', '', value.casefold())

    catalogue = await synchronise_catalogue()

    addon_names_to_catalogue_entries = bucketise(
        catalogue.entries, key=lambda i: normalise(i.name)
    )
    matches = (
        (a, addon_names_to_catalogue_entries.get(normalise(a.name))) for a in sorted(leftovers)
    )
    return [([a], uniq(Defn(i.source, i.id) for i in m)) for a, m in matches if m]


# In order of increasing heuristicitivenessitude
DEFAULT_MATCHERS: Mapping[str, Matcher] = {
    'toc_source_ids': match_toc_source_ids,
    'folder_name_subsets': match_folder_name_subsets,
    'folder_hashes': match_folder_hashes,
    'addon_names_with_folder_names': match_addon_names_with_folder_names,
}



from __future__ import annotations

import re

import pytest
from yarl import URL

from instawow import pkg_management
from instawow._sources.cfcore import CfCoreResolver
from instawow.definitions import Defn, Strategies, Strategy
from instawow.pkg_models import Pkg
from instawow.results import PkgFilesNotMatching
from instawow.shared_ctx import ConfigBoundCtx
from instawow.wow_installations import Flavour

CURSE_IDS = {
    'big-wigs': '2382',
    'molinari': '20338',
    'adibags': '23350',
    'mythic-dungeon-tools': '288981',
    'masque': '13592',
    'elkbuffbars': '2398',
    'atlaslootclassic': '326516',
    'elvui-adibags': '333072',
    'bigwigs-voice-korean': '402180',
}


@pytest.fixture
def curse_resolver(
    iw_config_ctx: ConfigBoundCtx,
):
    return CfCoreResolver(iw_config_ctx.config)


@pytest.mark.parametrize(
    'iw_profile_config_values',
    Flavour,
    indirect=True,
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_resolve_flavourful_addon(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['masque'])

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg


@pytest.mark.parametrize(
    'iw_profile_config_values',
    Flavour,
    indirect=True,
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_resolve_classic_only_addon(
    iw_config_ctx: ConfigBoundCtx,
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['atlaslootclassic'])

    result = (await curse_resolver.resolve([defn]))[defn]

    match iw_config_ctx.config.game_flavour:
        case Flavour.VanillaClassic | Flavour.WrathClassic:
            assert type(result) is Pkg
        case _:
            assert type(result) is PkgFilesNotMatching
            assert (
                result.message
                == f'no files found for: {Strategy.AnyFlavour}=None; {Strategy.AnyReleaseType}=None; {Strategy.VersionEq}=None'
            )


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_curse_any_flavour_strategy(
    curse_resolver: CfCoreResolver,
):
    flavourful = Defn(
        'curse', CURSE_IDS['masque'], strategies=Strategies({Strategy.AnyFlavour: True})
    )
    classics_only = Defn(
        'curse', CURSE_IDS['atlaslootclassic'], strategies=Strategies({Strategy.AnyFlavour: True})
    )

    results = await curse_resolver.resolve([flavourful, classics_only])
    assert all(type(r) is Pkg for r in results.values())


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_curse_slug_match(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', 'molinari')

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert result.id == CURSE_IDS['molinari']


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_curse_version_pinning(
    curse_resolver: CfCoreResolver,
):
    defn = Defn(
        'curse', 'molinari', strategies=Strategies({Strategy.VersionEq: '100005.97-Release'})
    )

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert result.options.version_eq is True
    assert result.version == '100005.97-Release'


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_curse_deps_retrieved(
    iw_config_ctx: ConfigBoundCtx,
):
    defn = Defn('curse', CURSE_IDS['bigwigs-voice-korean'])

    results = await pkg_management.resolve(iw_config_ctx, [defn], with_deps=True)
    assert all(type(r) is Pkg for r in results.values())
    assert {'bigwigs-voice-korean', 'big-wigs'} == {p.slug for p in results.values()}


@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_changelog_url_format(
    curse_resolver: CfCoreResolver,
):
    defn = Defn('curse', CURSE_IDS['masque'])

    result = (await curse_resolver.resolve([defn]))[defn]
    assert type(result) is Pkg
    assert re.match(
        r'https://api\.curseforge\.com/v1/mods/\d+/files/\d+/changelog',
        result.changelog_url,
    )


@pytest.mark.parametrize(
    ('url', 'extracted_alias'),
    [
        (
            'https://www.curseforge.com/wow/addons/molinari',
            'molinari',
        ),
        (
            'https://www.curseforge.com/wow/addons/molinari/download',
            'molinari',
        ),
    ],
)
def test_can_extract_alias_from_url(
    curse_resolver: CfCoreResolver,
    url: str,
    extracted_alias: str,
):
    assert curse_resolver.get_alias_from_url(URL(url)) == extracted_alias



from __future__ import annotations

import sys
from collections.abc import Callable, Hashable, Iterable, Iterator, Set
from itertools import chain, groupby, islice, repeat
from typing import TypeVar
from weakref import WeakValueDictionary

_T = TypeVar('_T')
_U = TypeVar('_U')
_THashable = TypeVar('_THashable', bound=Hashable)


class WeakValueDefaultDictionary(WeakValueDictionary[_T, _U]):
    def __init__(self, default_factory: Callable[[], _U]) -> None:
        super().__init__()
        self.__default_factory = default_factory

    def __getitem__(self, key: _T) -> _U:
        try:
            return super().__getitem__(key)
        except KeyError:
            default = self[key] = self.__default_factory()
            return default


def all_eq(it: Iterable[object]) -> bool:
    "Check that all elements of an iterable are equal."
    groups = groupby(it)
    return next(groups, True) and not next(groups, False)


if sys.version_info >= (3, 12):
    from itertools import batched as batched
else:

    def batched(iterable: Iterable[_T], n: int) -> Iterator[tuple[_T, ...]]:
        # batched('ABCDEFG', 3) → ABC DEF G
        if n < 1:
            raise ValueError('n must be at least one')
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch


def bucketise(iterable: Iterable[_U], key: Callable[[_U], _T]) -> dict[_T, list[_U]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket = dict[_T, list[_U]]()
    for value in iterable:
        bucket.setdefault(key(value), []).append(value)

    return bucket


def fill(it: Iterable[_T], fill: _T, length: int) -> Iterable[_T]:
    "Fill an iterable of specified length."
    return islice(chain(it, repeat(fill)), 0, length)


def merge_intersecting_sets(it: Iterable[Set[_T]]) -> Iterator[frozenset[_T]]:
    "Recursively merge intersecting sets in a collection."
    many_sets = list(map(frozenset, it))
    while many_sets:
        this_set = many_sets.pop(0)
        while True:
            for idx, other_set in enumerate(many_sets):
                if not this_set.isdisjoint(other_set):
                    this_set |= many_sets.pop(idx)
                    break
            else:
                break
        yield this_set


def uniq(it: Iterable[_THashable]) -> list[_THashable]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Iterable
from functools import wraps
from typing import Any, ParamSpec, TypeVar

_U = TypeVar('_U')
_P = ParamSpec('_P')


async def gather(it: Iterable[Awaitable[_U]]) -> list[_U]:
    return await asyncio.gather(*it)


def run_in_thread(fn: Callable[_P, _U]) -> Callable[_P, Awaitable[_U]]:
    @wraps(fn)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs):
        return asyncio.to_thread(fn, *args, **kwargs)

    return wrapper


async def cancel_tasks(tasks: Iterable[asyncio.Task[Any]]) -> None:
    incomplete_tasks = [t for t in tasks if not t.done()]
    for task in incomplete_tasks:
        task.cancel()
    await asyncio.gather(*incomplete_tasks, return_exceptions=True)



from __future__ import annotations

import pytest

from instawow._sources import DEFAULT_RESOLVERS
from instawow.shared_ctx import ConfigBoundCtx


def test_auth_bound_resolvers_are_not_unloaded_if_tokens_set(
    iw_config_ctx: ConfigBoundCtx,
):
    assert {
        r.metadata.id for r in DEFAULT_RESOLVERS if r.requires_access_token is not None
    }.issubset(iw_config_ctx.resolvers)


@pytest.mark.parametrize(
    'iw_global_config_values',
    [None],
    indirect=True,
)
def test_auth_bound_resolvers_are_unloaded_if_tokens_unset(
    iw_config_ctx: ConfigBoundCtx,
):
    assert {
        r.metadata.id for r in DEFAULT_RESOLVERS if r.requires_access_token is not None
    }.isdisjoint(iw_config_ctx.resolvers)



from __future__ import annotations

import re
from collections.abc import AsyncIterator, Sequence
from datetime import datetime, timedelta, timezone
from itertools import takewhile
from typing import Literal

from typing_extensions import NotRequired as N
from typing_extensions import TypedDict
from yarl import URL

from .. import pkg_models, shared_ctx
from .. import results as R
from .._logging import logger
from .._progress_reporting import make_default_progress
from .._utils.aio import gather
from .._utils.iteration import uniq
from .._utils.text import slugify
from .._utils.web import as_plain_text_data_url
from ..catalogue.cataloguer import CatalogueEntry
from ..definitions import ChangelogFormat, Defn, SourceMetadata
from ..resolvers import BaseResolver, PkgCandidate
from ..wow_installations import Flavour, FlavourVersionRange

_lock_prefix = object()

_LOAD_WOWI_CATALOGUE_LOCK = '_LOAD_WOWI_CATALOGUE_LOCK_'


class _WowiCommonTerms(TypedDict):
    UID: str  # Unique add-on ID
    UICATID: str  # ID of category add-on is placed in
    UIVersion: str  # Add-on version
    UIDate: int  # Upload date expressed as unix epoch
    UIName: str  # User-facing add-on name
    UIAuthorName: str


class _WowiListApiItem_CompatibilityEntry(TypedDict):
    version: str  # Game version, e.g. '8.3.0'
    name: str  # Xpac or patch name, e.g. "Visions of N'Zoth" for 8.3.0


class _WowiListApiItem(_WowiCommonTerms):
    UIFileInfoURL: str  # Add-on page on WoWI
    UIDownloadTotal: str  # Total number of downloads
    UIDownloadMonthly: str  # Number of downloads in the last month and not 'monthly'
    UIFavoriteTotal: str
    UICompatibility: list[_WowiListApiItem_CompatibilityEntry] | None  # ``null`` if would be empty
    UIDir: list[str]  # Names of folders contained in archive
    UIIMG_Thumbs: list[str] | None  # Thumbnail URLs; ``null`` if would be empty
    UIIMGs: list[str] | None  # Full-size image URLs; ``null`` if would be empty
    # There are only two add-ons on the entire list with siblings
    # (they refer to each other). I don't know if this was meant to capture
    # dependencies (probably not) but it's so underused as to be worthless.
    # ``null`` if would be empty
    UISiblings: list[str] | None
    UIDonationLink: N[str | None]  # Absent from the first item on the list (!)


class _WowiDetailsApiItem(_WowiCommonTerms):
    UIMD5: str | None  # Archive hash, ``null` when review pending
    UIFileName: str  # The actual filename, e.g. 'foo.zip'
    UIDownload: str  # Download URL
    UIPending: Literal['0', '1']  # Set to '1' if the file is awaiting approval
    UIDescription: str  # Long description with BB Code and all
    UIChangeLog: str  # This can also contain BB Code
    UIHitCount: str  # Same as UIDownloadTotal
    UIHitCountMonthly: str  # Same as UIDownloadMonthly


class _WowiCombinedItem(_WowiListApiItem, _WowiDetailsApiItem):
    pass


def _timestamp_to_datetime(timestamp: int):
    return datetime.fromtimestamp(timestamp / 1000, timezone.utc)


class WowiResolver(BaseResolver):
    metadata = SourceMetadata(
        id='wowi',
        name='WoWInterface',
        strategies=frozenset(),
        changelog_format=ChangelogFormat.Raw,
        addon_toc_key='X-WoWI-ID',
    )
    requires_access_token = None

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the as yet unreleased Minion v4,
    # which is fair to assume is unstable.  They changed the naming scheme to
    # camelCase and some fields which were strings were converted to numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # 'UICompatibility' can't be relied on to enforce compatibility
    # in instawow.  The API appears to inherit the version of the latest
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever the download link always points to the
    # 'retail' version, which for single-file add-ons belonging to the
    # classic category would be an add-on for classic.
    __list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    __details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    @classmethod
    def get_alias_from_url(cls, url: URL) -> str | None:
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                return url.query.get('fileid')
            elif url.name == 'fileinfo.php':
                return url.query.get('id')
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                return match and match['id']

    async def __synchronise(self):
        async with shared_ctx.locks[_lock_prefix, _LOAD_WOWI_CATALOGUE_LOCK]:
            async with shared_ctx.web_client.get(
                self.__list_api_url,
                expire_after=timedelta(hours=1),
                raise_for_status=True,
                trace_request_ctx={
                    'progress': make_default_progress(
                        type_='download',
                        label=f'Synchronising {self.metadata.name} catalogue',
                    )
                },
            ) as response:
                list_items_by_id: dict[str, _WowiListApiItem] = {
                    i['UID']: i for i in await response.json()
                }
                return list_items_by_id

    async def resolve(
        self, defns: Sequence[Defn]
    ) -> dict[Defn, pkg_models.Pkg | R.ManagerError | R.InternalError]:
        list_items_by_id = await self.__synchronise()

        defns_to_ids = {d: ''.join(takewhile(str.isdigit, d.alias)) for d in defns}
        async with shared_ctx.web_client.get(
            (
                self.__details_api_url
                / f'{",".join(uniq(i for i in defns_to_ids.values() if i))}.json'
            ),
            expire_after=timedelta(minutes=5),
        ) as response:
            if response.status == 404:
                return await super().resolve(defns)

            response.raise_for_status()

            details_items_by_id: dict[str, _WowiDetailsApiItem] = {
                i['UID']: i for i in await response.json()
            }

        results = await gather(
            R.resultify_async_exc(self.resolve_one(d, {**a, **b} if a and b else None))
            for d, i in defns_to_ids.items()
            for a, b in ((list_items_by_id.get(i), details_items_by_id.get(i)),)
        )
        return dict(zip(defns, results))

    async def _resolve_one(self, defn: Defn, metadata: _WowiCombinedItem | None) -> PkgCandidate:
        if metadata is None:
            raise R.PkgNonexistent

        return PkgCandidate(
            id=metadata['UID'],
            slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=metadata['UIFileInfoURL'],
            download_url=metadata['UIDownload'],
            date_published=_timestamp_to_datetime(metadata['UIDate']),
            version=metadata['UIVersion'],
            changelog_url=as_plain_text_data_url(metadata['UIChangeLog']),
        )

    @classmethod
    async def catalogue(cls) -> AsyncIterator[CatalogueEntry]:
        logger.debug(f'retrieving {cls.__list_api_url}')

        async with shared_ctx.web_client.get(
            cls.__list_api_url, raise_for_status=True
        ) as response:
            items: list[_WowiListApiItem] = await response.json()

        for item in items:
            match item:
                case {'UICATID': '160'}:
                    game_flavours = (Flavour.VanillaClassic,)
                case {'UICATID': '161'}:
                    # TBC Classic
                    continue
                case {'UICATID': '162'}:
                    game_flavours = (Flavour.Classic,)
                case _:
                    compatibility = item['UICompatibility']
                    if compatibility is None or len(compatibility) < 2:
                        game_flavours = (Flavour.Retail,)
                    else:
                        game_flavours = (
                            Flavour.from_flavour_keyed_enum(r)
                            for c in compatibility
                            if (r := FlavourVersionRange.from_version(c['version']))
                        )

            yield CatalogueEntry(
                source=cls.metadata.id,
                id=item['UID'],
                name=item['UIName'],
                url=item['UIFileInfoURL'],
                game_flavours=frozenset(game_flavours),
                download_count=int(item['UIDownloadTotal']),
                last_updated=_timestamp_to_datetime(item['UIDate']),
                folders=[frozenset(item['UIDir'])],
            )



from __future__ import annotations

from collections.abc import Iterable
from functools import lru_cache
from typing import Protocol

import click
import pluggy

from . import resolvers

_project_name = __spec__.parent
_entry_point = f'{_project_name}.plugins'

hookspec = pluggy.HookspecMarker(_project_name)
hookimpl = pluggy.HookimplMarker(_project_name)


class InstawowPlugin(Protocol):  # pragma: no cover
    "The plug-in interface."

    @hookspec
    def instawow_add_commands(self) -> Iterable[click.Command]:
        "Additional commands to register with ``click``."
        ...

    @hookspec
    def instawow_add_resolvers(self) -> Iterable[type[resolvers.Resolver]]:
        "Additional resolvers to load."
        ...


@lru_cache(1)
def _load_plugins():
    plugin_manager = pluggy.PluginManager(_project_name)
    plugin_manager.add_hookspecs(InstawowPlugin)
    plugin_manager.load_setuptools_entrypoints(_entry_point)
    return plugin_manager.hook


def get_plugin_commands() -> Iterable[Iterable[click.Command]]:
    plugin_hook = _load_plugins()
    return plugin_hook.instawow_add_commands()


def get_plugin_resolvers() -> Iterable[Iterable[type[resolvers.Resolver]]]:
    plugin_hook = _load_plugins()
    return plugin_hook.instawow_add_resolvers()



# pyright: strict

from __future__ import annotations

import importlib.resources
import json
import re
from collections.abc import Awaitable, Callable
from functools import cache
from io import BytesIO
from typing import Any
from zipfile import ZipFile

import attrs
from aiohttp.web import Request, Response
from yarl import URL

from instawow import __version__

_match_any = re.compile(r'.*')


def _load_fixture(filename: str):
    return (importlib.resources.files(__spec__.parent) / filename).read_bytes()


def _load_json_fixture(filename: str):
    return json.loads(_load_fixture(filename))


@cache
def _make_addon_zip(*folders: str):
    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        for folder in folders:
            file.writestr(f'{folder}/{folder}.toc', b'')

    return buffer.getvalue()


@attrs.frozen
class Route:
    url: URL = attrs.field(converter=URL)
    response: (
        Response
        | Callable[[Request], Response]
        | Callable[[Request], Awaitable[Response]]
        | dict[str, Any]
        | str
    )
    path_pattern: re.Pattern[str] | None = None
    method: str = 'GET'
    body_pattern: re.Pattern[str] | None = None
    match_querystring: bool = False
    repeat: float = float('inf')
    case_insensitive: bool = False

    def _make_path_pattern(self):
        if self.path_pattern is not None:
            return self.path_pattern

        if self.case_insensitive:
            return re.compile(rf'^{re.escape(self.url.path_qs)}$', re.IGNORECASE)

        return self.url.path_qs

    def to_aresponses_add_args(self) -> dict[str, Any]:
        return {
            'host_pattern': self.url.host,
            'path_pattern': self._make_path_pattern(),
            'method_pattern': self.method,
            'body_pattern': _match_any if self.body_pattern is None else self.body_pattern,
            'match_querystring': self.match_querystring,
            'repeat': self.repeat,
            'response': self.response,
        }


def _make_route_dict_entry(route: Route):
    return (route.url, route)


ROUTES = dict(
    map(
        _make_route_dict_entry,
        [
            Route(
                '//pypi.org/pypi/instawow/json',
                {'info': {'version': __version__}},
            ),
            Route(
                '//raw.githubusercontent.com/layday/instawow-data/data/base-catalogue-v7.compact.json',
                _load_json_fixture('base-catalogue-v7.compact.json'),
            ),
            Route(
                '//api.curseforge.com/v1/mods',
                _load_json_fixture('curse-addon--all.json'),
                method='POST',
            ),
            Route(
                '//api.curseforge.com/v1/mods/search?gameId=1&slug=molinari',
                _load_json_fixture('curse-addon-slug-search.json'),
                match_querystring=True,
            ),
            Route(
                '//api.curseforge.com/v1/mods/20338/files',
                _load_json_fixture('curse-addon-files.json'),
            ),
            Route(
                '//api.curseforge.com/v1/mods/20338/files/{id}/changelog',
                _load_json_fixture('curse-addon-changelog.json'),
                path_pattern=re.compile(r'^/v1/mods/20338/files/(\d+)/changelog$'),
            ),
            Route(
                '//edge.forgecdn.net',
                lambda _: Response(body=_make_addon_zip('Molinari')),
                path_pattern=_match_any,
            ),
            Route(
                '//api.mmoui.com/v3/game/WOW/filelist.json',
                _load_json_fixture('wowi-filelist.json'),
            ),
            Route(
                '//api.mmoui.com/v3/game/WOW/filedetails/{id}.json',
                _load_json_fixture('wowi-filedetails.json'),
                path_pattern=re.compile(r'^/v3/game/WOW/filedetails/(\d*)\.json$'),
            ),
            Route(
                '//cdn.wowinterface.com',
                lambda _: Response(body=_make_addon_zip('Molinari')),
                path_pattern=_match_any,
            ),
            Route(
                '//api.tukui.org/v1/addon/tukui',
                _load_json_fixture('tukui-ui--tukui.json'),
            ),
            Route(
                '//api.tukui.org/v1/addon/elvui',
                _load_json_fixture('tukui-ui--elvui.json'),
            ),
            Route(
                '//api.tukui.org/v1/download/',
                lambda _: Response(body=_make_addon_zip('Tukui')),
                path_pattern=re.compile(r'^/v1/download/'),
            ),
            Route(
                '//api.github.com/repos/nebularg/PackagerTest',
                _load_json_fixture('github-repo-release-json.json'),
            ),
            Route(
                '//api.github.com/repos/nebularg/PackagerTest/releases?per_page=10',
                _load_json_fixture('github-release-release-json.json'),
                match_querystring=True,
            ),
            Route(
                '//api.github.com/repos/nebularg/PackagerTest/releases/assets/37156458',
                _load_json_fixture('github-release-release-json-release-json.json'),
            ),
            Route(
                '//api.github.com/repositories/388670',
                _load_json_fixture('github-repo-molinari.json'),
            ),
            Route(
                '//api.github.com/repos/p3lim-wow/Molinari',
                _load_json_fixture('github-repo-molinari.json'),
                case_insensitive=True,
            ),
            Route(
                '//api.github.com/repositories/388670/releases?per_page=10',
                _load_json_fixture('github-release-molinari.json'),
                case_insensitive=True,
                match_querystring=True,
            ),
            Route(
                '//api.github.com/repos/p3lim-wow/Molinari/releases?per_page=10',
                _load_json_fixture('github-release-molinari.json'),
                case_insensitive=True,
                match_querystring=True,
            ),
            Route(
                URL(
                    next(
                        a['url']
                        for a in _load_json_fixture('github-release-molinari.json')[0]['assets']
                        if a['name'] == 'release.json'
                    )
                ).with_scheme(''),
                _load_json_fixture('github-release-molinari-release-json.json'),
                case_insensitive=True,
            ),
            Route(
                '//api.github.com/repos/AdiAddons/AdiBags',
                _load_json_fixture('github-repo-no-releases.json'),
            ),
            Route(
                '//api.github.com/repos/AdiAddons/AdiBags/releases?per_page=10',
                lambda _: Response(body=b'', status=404),
                match_querystring=True,
            ),
            Route(
                '//api.github.com/repos/AdiAddons/AdiButtonAuras/releases/tags/2.0.19',
                _load_json_fixture('github-release-no-assets.json'),
            ),
            Route(
                '//api.github.com/repos/layday/foobar',
                lambda _: Response(body=b'', status=404),
            ),
            Route(
                '//api.github.com/repos/{x}/{y}/releases/asssets/{z}',
                lambda _: Response(body=_make_addon_zip('Molinari')),
                path_pattern=re.compile(r'^/repos(/[^/]*){2}/releases/assets/'),
            ),
            Route(
                '//github.com/login/device/code',
                _load_json_fixture('github-oauth-login-device-code.json'),
                method='POST',
            ),
            Route(
                '//github.com/login/oauth/access_token',
                _load_json_fixture('github-oauth-login-access-token.json'),
                method='POST',
            ),
            Route(
                '//api.github.com/repos/28/NoteworthyII',
                _load_json_fixture('github-repo-no-release-json.json'),
            ),
            Route(
                '//api.github.com/repos/28/NoteworthyII/releases?per_page=10',
                _load_json_fixture('github-release-no-release-json.json'),
                match_querystring=True,
            ),
            Route(
                '//addons.wago.io/api/external/addons/_match',
                _load_json_fixture('wago-match-addons.json'),
                method='POST',
            ),
        ],
    )
)



from __future__ import annotations

from collections.abc import Iterator, Mapping
from functools import cached_property
from pathlib import Path

from typing_extensions import Self


class TocReader(Mapping[str, str]):
    """Extracts key-value pairs from TOC files."""

    def __init__(self, contents: str) -> None:
        self._entries = {
            k: v
            for e in contents.splitlines()
            if e.startswith('##')
            for k, v in (map(str.strip, e.lstrip('#').partition(':')[::2]),)
            if k
        }

    def __iter__(self) -> Iterator[str]:
        return iter(self._entries)

    def __getitem__(self, key: str, /) -> str:
        return self._entries[key]

    def __len__(self) -> int:
        return len(self._entries)

    @classmethod
    def from_path(cls, path: Path) -> Self:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'))

    @cached_property
    def interfaces(self) -> list[int]:
        return [int(i) for i in self.get('Interface', '').split(',') if i]

    @cached_property
    def version(self) -> str:
        return self.get('Version', '')



from __future__ import annotations

from .cfcore import CfCoreResolver
from .github import GithubResolver
from .instawow import InstawowResolver
from .tukui import TukuiResolver
from .wago import WagoResolver
from .wowi import WowiResolver

DEFAULT_RESOLVERS = [
    GithubResolver,
    CfCoreResolver,
    WowiResolver,
    TukuiResolver,
    InstawowResolver,
    WagoResolver,
]



from __future__ import annotations

import datetime as dt
import enum
import textwrap
from collections.abc import Awaitable, Callable, Collection, Iterable, Mapping, Sequence
from functools import cached_property, partial, reduce
from itertools import chain, count, repeat
from typing import TYPE_CHECKING, Any, Generic, NoReturn, TypeVar, overload

import attrs
import click
import click.types
from typing_extensions import Self

from . import __version__, _logging, pkg_management, pkg_models, shared_ctx
from . import config as _config
from . import results as R
from ._utils.compat import StrEnum
from ._utils.iteration import all_eq, bucketise, uniq
from .definitions import ChangelogFormat, Defn, Strategy
from .plugins import get_plugin_commands

_T = TypeVar('_T')
_TStrEnum = TypeVar('_TStrEnum', bound=StrEnum)


class Report:
    SUCCESS_SYMBOL = click.style('✓', fg='green')
    FAILURE_SYMBOL = click.style('✗', fg='red')
    WARNING_SYMBOL = click.style('!', fg='blue')

    def __init__(
        self,
        results: Iterable[tuple[Defn, R.Result]],
        filter_fn: Callable[[R.Result], bool] = lambda _: True,
    ) -> None:
        self.results = list(results)
        self.filter_fn = filter_fn

    @property
    def exit_code(self) -> int:
        return any(
            isinstance(r, R.ManagerError | R.InternalError) and self.filter_fn(r)
            for _, r in self.results
        )

    @classmethod
    def _result_type_to_symbol(cls, result: R.Result) -> str:
        match result:
            case R.InternalError():
                return cls.WARNING_SYMBOL
            case R.ManagerError():
                return cls.FAILURE_SYMBOL
            case _:
                return cls.SUCCESS_SYMBOL

    def __str__(self) -> str:
        return '\n'.join(
            f'{self._result_type_to_symbol(r)} {click.style(a.as_uri(), bold=True)}\n'
            f'{textwrap.fill(r.message, initial_indent=" " * 2, subsequent_indent=" " * 4)}'
            for a, r in self.results
            if self.filter_fn(r)
        )

    def generate(self) -> None:
        config_ctx: ConfigBoundCtxProxy | None = click.get_current_context().obj
        if config_ctx and config_ctx.config.global_config.auto_update_check:
            from ._version_check import is_outdated

            outdated, new_version = run_with_progress(is_outdated(config_ctx.config.global_config))
            if outdated:
                click.echo(f'{self.WARNING_SYMBOL} instawow v{new_version} is available')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> NoReturn:
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.exit_code)


if TYPE_CHECKING:
    ConfigBoundCtxProxyBase = shared_ctx.ConfigBoundCtx
else:
    ConfigBoundCtxProxyBase = object


class ConfigBoundCtxProxy(ConfigBoundCtxProxyBase):
    def __init__(self, click_ctx: click.Context) -> None:
        self.__click_ctx = click_ctx

    if not TYPE_CHECKING:

        @cached_property
        def __config_ctx(self) -> shared_ctx.ConfigBoundCtx:
            global_config = _config.GlobalConfig.read().ensure_dirs()
            try:
                config = _config.ProfileConfig.read(
                    global_config, self.__click_ctx.params['profile']
                ).ensure_dirs()
            except FileNotFoundError:
                config = self.__click_ctx.invoke(configure)

            _logging.setup_logging(config.logging_dir, *self.__click_ctx.params['verbose'])

            return self.__click_ctx.with_resource(shared_ctx.ConfigBoundCtx(config))

        def __getattr__(self, name: str) -> Any:
            return getattr(self.__config_ctx, name)


def run_with_progress(awaitable: Awaitable[_T], click_ctx: click.Context | None = None) -> _T:
    import asyncio

    from .http import init_web_client

    if click_ctx is None:
        click_ctx = click.get_current_context()
    click_ctx = click_ctx.find_root()

    config_ctx: ConfigBoundCtxProxy = click_ctx.obj

    make_init_web_client = partial(
        init_web_client,
        config_ctx.config.global_config.http_cache_dir,
        no_cache=click_ctx.params['no_cache'],
    )

    if any(click_ctx.params['verbose']):

        async def run():
            async with make_init_web_client() as web_client:
                shared_ctx.web_client_var.set(web_client)
                return await awaitable

    else:
        from ._cli_prompts import ProgressBar, make_progress_bar_group
        from ._progress_reporting import make_progress_receiver
        from ._utils.aio import cancel_tasks

        async def run():
            with (
                make_progress_receiver[pkg_management.PkgDownloadProgress]() as iter_progress,
                make_progress_bar_group() as progress_bar_group,
            ):

                async def observe_progress():
                    progress_bars = dict[int, ProgressBar]()

                    try:
                        async for progress_group in iter_progress:
                            for progress_id in progress_bars.keys() - progress_group.keys():
                                del progress_bars[progress_id]

                            for progress_id, progress in (
                                (k, progress_group[k])
                                for k in progress_bars.keys() ^ progress_group.keys()
                            ):
                                match progress:
                                    case {'type_': 'pkg_download', 'defn': defn}:
                                        label = f'Downloading {defn.as_uri()}'
                                    case {'label': str(label)}:
                                        pass
                                    case _:
                                        continue

                                progress_bars[progress_id] = ProgressBar(
                                    progress_bar=progress_bar_group,
                                    label=label,
                                    total=progress['total'],
                                    is_download=progress.get('unit') == 'bytes',
                                )

                            progress_bar_group.counters = list(progress_bars.values())

                            for progress_id, progress in progress_group.items():
                                progress_bars[progress_id].items_completed = progress['current']
                                progress_bar_group.invalidate()

                    finally:
                        progress_bar_group.counters = []

                observe_progress_task = asyncio.create_task(observe_progress())

                try:
                    async with make_init_web_client(with_progress=True) as web_client:
                        shared_ctx.web_client_var.set(web_client)
                        return await awaitable

                finally:
                    await cancel_tasks([observe_progress_task])

    return asyncio.run(run())


def _with_obj(fn: Callable[..., object]):
    def wrapper(ctx: click.Context, __: click.Parameter, value: object):
        return fn(ctx.obj, value)

    return wrapper


class _StrEnumChoiceParam(Generic[_TStrEnum], click.Choice):
    def __init__(
        self,
        choice_enum: type[_TStrEnum],
        case_sensitive: bool = True,
    ) -> None:
        super().__init__(
            choices=list(choice_enum),
            case_sensitive=case_sensitive,
        )
        self.__choice_enum = choice_enum

    def convert(
        self, value: Any, param: click.Parameter | None, ctx: click.Context | None
    ) -> _TStrEnum:
        converted_value = super().convert(value, param, ctx)
        return self.__choice_enum(converted_value)


class _ManyOptionalChoiceValueParam(click.types.CompositeParamType):
    name = 'optional-choice-value'

    def __init__(
        self,
        choice_param: click.Choice,
        *,
        value_types: Mapping[str, click.types.ParamType] = {},
    ) -> None:
        super().__init__()
        self.__choice_param = choice_param
        self.__value_types = value_types

    def __parse_value(self, value: tuple[str, ...]):
        return (
            (k, v, self.__choice_param, vc)
            for r in value
            for k, _, v in (r.partition('='),)
            for vc in (self.__value_types.get(k),)
        )

    @property
    def arity(self) -> int:
        return -1

    def convert(
        self, value: tuple[str, ...], param: click.Parameter | None, ctx: click.Context | None
    ) -> Any:
        return {
            kc.convert(k, param, ctx): vc.convert(v, param, ctx) if vc and v else (v or None)
            for k, v, kc, vc in self.__parse_value(value)
        }

    def get_metavar(self, param: click.Parameter) -> str:
        return f'{{{",".join(self.__choice_param.choices)}}}[=VALUE]'


def _register_plugin_commands(group: click.Group):
    additional_commands = (c for g in get_plugin_commands() for c in g)
    for command in additional_commands:
        group.add_command(command)
    return group


def _parse_debug_option(
    _: click.Context, __: click.Parameter, value: float
) -> tuple[bool, bool, bool]:
    return (value > 0, value > 1, value > 2)


class _SectionedHelpGroup(click.Group):
    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        command_sections = bucketise(
            ((s, c) for s, c in self.commands.items() if not c.hidden),
            key=lambda c: 'Command groups' if isinstance(c[1], click.Group) else 'Commands',
        )
        if command_sections:
            for section_name, commands in command_sections.items():
                with formatter.section(section_name):
                    limit = formatter.width - 6 - max(len(s) for s, _ in commands)
                    formatter.write_dl(
                        [(s, c.get_short_help_str(limit)) for s, c in commands],
                    )


@overload
def _parse_uri(
    config_ctx: ConfigBoundCtxProxy,
    value: str,
    *,
    raise_invalid: bool = True,
) -> Defn: ...


@overload
def _parse_uri(
    config_ctx: ConfigBoundCtxProxy,
    value: list[str],
    *,
    raise_invalid: bool = True,
) -> list[Defn]: ...


def _parse_uri(
    config_ctx: ConfigBoundCtxProxy,
    value: str | list[str] | None,
    *,
    raise_invalid: bool = True,
) -> Defn | list[Defn] | None:
    if value is None:
        return None

    if not isinstance(value, str):
        defns = (_parse_uri(config_ctx, v, raise_invalid=raise_invalid) for v in value)
        return uniq(defns)

    try:
        defn = Defn.from_uri(value, known_sources=config_ctx.resolvers, allow_unsourced=True)
    except ValueError as exc:
        raise click.BadParameter(exc.args[0]) from None

    if not defn.source:
        match = pkg_management.get_alias_from_url(config_ctx, defn.alias)
        if match:
            source, alias = match
            defn = attrs.evolve(defn, source=source, alias=alias)
        elif raise_invalid and ':' not in defn.alias:
            raise click.BadParameter(value)

    return defn


def _make_pkg_where_clause_and_params(defns: Sequence[Defn]) -> tuple[str, dict[str, Any]]:
    def make_inner():
        iter_named_param = (f'where_param_{i}' for i in count())
        for defn, source_param, alias_param in zip(defns, iter_named_param, iter_named_param):
            if not defn.source:
                yield (
                    f"(pkg.slug LIKE '%' || :{alias_param} || '%')",
                    {alias_param: defn.alias},
                )
            elif not defn.alias:
                yield (
                    f'pkg.source = :{source_param}',
                    {source_param: defn.source},
                )
            else:
                yield (
                    f'pkg.source = :{source_param} AND (pkg.id = :{alias_param} OR lower(pkg.slug) = lower(:{alias_param}))',
                    {source_param: defn.source, alias_param: defn.alias},
                )

    if defns:
        where_clauses, where_params = zip(*make_inner())
        return (
            'WHERE\n  ' + '\n  OR '.join(where_clauses),
            reduce(lambda a, b: a | b, where_params, {}),
        )
    else:
        return ('', {})


class _ListFormat(StrEnum):
    Simple = enum.auto()
    Detailed = enum.auto()
    Json = enum.auto()


@click.group(context_settings={'help_option_names': ('-h', '--help')}, cls=_SectionedHelpGroup)
@click.version_option(__version__, prog_name=__spec__.parent)
@click.option(
    '--verbose',
    '-v',
    count=True,
    help='Log incrementally more things.  Additive.',
    callback=_parse_debug_option,
)
@click.option(
    '--no-cache',
    is_flag=True,
    default=False,
    help='Disable the HTTP cache.',
)
@click.option(
    '--profile',
    '-p',
    default='__default__',
    help='Activate the specified profile.',
)
@click.pass_context
def cli(ctx: click.Context, **__: object) -> None:
    "Add-on manager for World of Warcraft."
    ctx.obj = ConfigBoundCtxProxy(ctx)


@cli.command
@click.argument('addons', nargs=-1, callback=_with_obj(_parse_uri))
@click.option(
    '--replace',
    'replace_folders',
    is_flag=True,
    default=False,
    help='Replace unreconciled add-ons.',
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to install add-ons.  Add-on archives will not be download and the '
    'database will not be modified.',
)
@click.pass_obj
def install(
    config_ctx: ConfigBoundCtxProxy,
    addons: Sequence[Defn],
    replace_folders: bool,
    dry_run: bool,
) -> None:
    "Install add-ons."
    results = run_with_progress(
        pkg_management.install(
            config_ctx, addons, replace_folders=replace_folders, dry_run=dry_run
        )
    )
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, callback=_with_obj(_parse_uri))
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help='Pretend to update add-ons.  Add-on archives will not be download and the '
    'database will not be modified.  Use this option to check for updates.',
)
@click.pass_obj
def update(
    config_ctx: ConfigBoundCtxProxy,
    addons: Sequence[Defn],
    dry_run: bool,
) -> None:
    "Update installed add-ons."

    def filter_results(result: R.Result):
        # Hide packages from output if they are up to date and not pinned.
        return True if addons or not isinstance(result, R.PkgUpToDate) else result.is_pinned

    results = run_with_progress(
        pkg_management.update(config_ctx, addons or 'all', dry_run=dry_run)
    )
    Report(results.items(), filter_results).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, required=True, callback=_with_obj(_parse_uri))
@click.option(
    '--keep-folders',
    is_flag=True,
    default=False,
    help='Remove the add-on from the database but do not delete its folders.',
)
@click.pass_obj
def remove(config_ctx: ConfigBoundCtxProxy, addons: Sequence[Defn], keep_folders: bool) -> None:
    "Remove add-ons."
    results = run_with_progress(
        pkg_management.remove(config_ctx, addons, keep_folders=keep_folders)
    )
    Report(results.items()).generate_and_exit()


@cli.command
@click.argument('addon', callback=_with_obj(_parse_uri))
@click.option(
    '--undo',
    is_flag=True,
    default=False,
    help='Undo rollback by reinstalling an add-on using the default strategy.',
)
@click.pass_obj
def rollback(config_ctx: ConfigBoundCtxProxy, addon: Defn, undo: bool) -> None:
    "Roll an add-on back to an older version."
    from ._cli_prompts import Choice, select_one

    pkg = pkg_management.get_pkg(config_ctx, addon)
    if not pkg:
        Report([(addon, R.PkgNotInstalled())]).generate_and_exit()
    elif Strategy.VersionEq not in config_ctx.resolvers[pkg.source].metadata.strategies:
        Report([(addon, R.PkgStrategiesUnsupported({Strategy.VersionEq}))]).generate_and_exit()
    elif undo:
        Report(
            run_with_progress(
                pkg_management.update(config_ctx, [addon.with_default_strategy_set()])
            ).items()
        ).generate_and_exit()

    reconstructed_defn = pkg.to_defn()

    versions = pkg.logged_versions
    if len(versions) <= 1:
        Report([(addon, R.PkgFilesMissing('cannot find older versions'))]).generate_and_exit()

    choices = [
        Choice(v.version, value=v.version, disabled=v.version == pkg.version) for v in versions
    ]
    selection = select_one(
        f'Select version of {reconstructed_defn.as_uri()} for rollback', choices
    ).prompt()

    Report(
        run_with_progress(
            pkg_management.update(config_ctx, [reconstructed_defn.with_version(selection)])
        ).items()
    ).generate_and_exit()


@cli.command
@click.option('--auto', '-a', is_flag=True, default=False, help='Do not ask for user input.')
@click.option(
    '--list-unreconciled', is_flag=True, default=False, help='List unreconciled add-ons and exit.'
)
@click.pass_obj
def reconcile(config_ctx: ConfigBoundCtxProxy, auto: bool, list_unreconciled: bool) -> None:
    "Reconcile pre-installed add-ons."
    from ._cli_prompts import SKIP, Choice, confirm, select_one
    from ._utils.text import tabulate
    from .matchers import DEFAULT_MATCHERS, AddonFolder, get_unreconciled_folders

    leftovers = get_unreconciled_folders(config_ctx)
    if list_unreconciled and leftovers:
        click.echo(tabulate([('unreconciled',), *((f.name,) for f in sorted(leftovers))]))
        return
    elif not leftovers:
        click.echo('No add-ons left to reconcile.')
        return

    if not auto:
        click.echo(
            textwrap.dedent(
                """\
                Use the arrow keys to navigate, <o> to open an add-on in your browser,
                enter to make a selection and <s> to skip to the next item.

                Versions that differ from the installed version or differ between
                choices are highlighted in purple.

                instawow will perform three passes in decreasing order of accuracy,
                looking to match source IDs and add-on names in TOC files, and folders.

                Selected add-ons will be reinstalled.

                You can also run `reconcile` in automatic mode by passing
                the `--auto` flag.  In this mode, add-ons will be reconciled
                without user input.
                """
            )
        )

    if auto:

        def confirm_install() -> bool:
            return True

        def select_pkg(
            addons: Sequence[AddonFolder], pkgs: Sequence[pkg_models.Pkg]
        ) -> Defn | None:
            return pkgs[0].to_defn()
    else:

        def confirm_install():
            return confirm('Install selected add-ons?').prompt()

        def select_pkg(addons: Sequence[AddonFolder], pkgs: Sequence[pkg_models.Pkg]):
            def construct_choice(pkg: pkg_models.Pkg, disabled: bool):
                defn = pkg.to_defn()
                return Choice(
                    [
                        ('', f'{defn.as_uri()}=='),
                        ('class:attention' if highlight_version else '', pkg.version),
                    ],
                    defn,
                    browser_url=pkg.url,
                    disabled=disabled,
                )

            # Highlight versions if they are disparate
            highlight_version = not all_eq(
                chain((a.toc_reader.version for a in addons), (p.version for p in pkgs))
            )

            selection = select_one(
                f'{textwrap.shorten(", ".join(a.name for a in addons), 60)}'
                f' [{addons[0].toc_reader.version or "?"}]',
                [construct_choice(p, False) for p in pkgs],
                can_skip=True,
            ).prompt()
            return selection if selection is not SKIP else None

    def gather_selections(
        groups: Collection[tuple[Sequence[AddonFolder], Sequence[Defn]]],
        pkgs: Mapping[Defn, pkg_models.Pkg],
    ):
        for addon_folder, defns in groups:
            shortlist = [p for d in defns for p in (pkgs.get(d),) if p]
            if shortlist:
                selection = select_pkg(addon_folder, shortlist)
                yield selection
            else:
                yield None

    for fn in DEFAULT_MATCHERS.values():
        defn_groups = run_with_progress(fn(config_ctx, leftovers))
        resolve_results = run_with_progress(
            pkg_management.resolve(config_ctx, uniq(d for _, b in defn_groups for d in b))
        )
        pkgs, _ = pkg_management.bucketise_results(resolve_results.items())

        selections = [s for s in gather_selections(defn_groups, pkgs) if s]
        if selections and confirm_install():
            results = run_with_progress(
                pkg_management.install(config_ctx, selections, replace_folders=True)
            )
            Report(results.items()).generate()

        leftovers = get_unreconciled_folders(config_ctx)
        if not leftovers:
            break

    if leftovers:
        click.echo()
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))


@cli.command
@click.argument('addons', nargs=-1, callback=_with_obj(partial(_parse_uri, raise_invalid=False)))
@click.pass_obj
def rereconcile(config_ctx: ConfigBoundCtxProxy, addons: Sequence[Defn]) -> None:
    "Rereconcile installed add-ons."
    from ._cli_prompts import SKIP, Choice, confirm, select_one

    def select_alternative_pkg(pkg: pkg_models.Pkg, equivalent_pkg_defns: Sequence[Defn]):
        def construct_choice(equivalent_pkg: pkg_models.Pkg, disabled: bool):
            defn = equivalent_pkg.to_defn()
            return Choice(
                [
                    ('', f'{defn.as_uri()}=='),
                    ('class:attention' if highlight_version else '', equivalent_pkg.version),
                ],
                defn,
                browser_url=equivalent_pkg.url,
                disabled=disabled,
            )

        shortlisted_pkgs = [p for d in equivalent_pkg_defns for p in (pkgs.get(d),) if p]
        if not shortlisted_pkgs:
            return None

        highlight_version = not all_eq(i.version for i in (pkg, *shortlisted_pkgs))

        selection = select_one(
            pkg.name,
            [
                construct_choice(pkg, True),
                *(construct_choice(p, False) for p in shortlisted_pkgs),
            ],
            can_skip=True,
        ).prompt()
        return selection if selection is not SKIP else None

    with config_ctx.database.connect() as connection:
        query = """
            SELECT *
            FROM pkg
            {where_clause}
            ORDER BY lower(name)
        """
        if addons:
            where_clause, query_params = _make_pkg_where_clause_and_params(addons)
            execute = partial(
                connection.execute, query.format(where_clause=where_clause), query_params
            )
        else:
            execute = partial(connection.execute, query.format(where_clause=''))

        installed_pkgs = [
            pkg_models.build_pkg_from_row_mapping(connection, p) for p in execute().fetchall()
        ]

    equivalent_pkg_defn_groups = run_with_progress(
        pkg_management.find_equivalent_pkg_defns(config_ctx, installed_pkgs)
    )
    resolve_results = run_with_progress(
        pkg_management.resolve(
            config_ctx, uniq(d for b in equivalent_pkg_defn_groups.values() for d in b)
        )
    )
    pkgs, _ = pkg_management.bucketise_results(resolve_results.items())

    selections = {
        p.to_defn(): s
        for p, d in equivalent_pkg_defn_groups.items()
        for s in (select_alternative_pkg(p, d),)
        if s
    }

    if selections and confirm('Install selected add-ons?').prompt():
        Report(
            run_with_progress(pkg_management.replace(config_ctx, selections)).items()
        ).generate_and_exit()


def _concat_search_terms(_: click.Context, __: click.Parameter, value: tuple[str, ...]):
    return ' '.join(value)


def _parse_iso_date_into_datetime(_: click.Context, __: click.Parameter, value: str | None):
    if value is not None:
        return dt.datetime.strptime(value, '%Y-%m-%d').replace(tzinfo=dt.timezone.utc)


@cli.command
@click.argument('search-terms', nargs=-1, required=True, callback=_concat_search_terms)
@click.option(
    '--limit',
    '-l',
    default=10,
    show_default=True,
    type=click.IntRange(1, 20, clamp=True),
    help='Maximum results to return.',
)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
@click.option(
    '--source',
    'sources',
    multiple=True,
    help='Search only in the specified source.  Multiple option.',
)
@click.option(
    '--prefer-source',
    help='Hide duplicates from other sources.',
)
@click.option(
    '--no-exclude-installed',
    is_flag=True,
    default=False,
    help='Do not exclude installed add-ons from search results.',
)
@click.pass_context
def search(
    ctx: click.Context,
    search_terms: str,
    limit: int,
    sources: Sequence[str],
    prefer_source: str | None,
    start_date: dt.datetime | None,
    no_exclude_installed: bool,
) -> None:
    "Search for add-ons to install."
    from ._cli_prompts import Choice, confirm, select_multiple
    from .catalogue.search import search

    catalogue_entries = run_with_progress(
        search(
            ctx.obj,
            search_terms,
            limit=limit,
            sources=frozenset(sources),
            prefer_source=prefer_source,
            start_date=start_date,
            filter_installed='ident' if no_exclude_installed else 'exclude_from_all_sources',
        ),
        ctx,
    )
    if catalogue_entries:
        choices = [
            Choice(f'{e.name}  ({d.as_uri()})', d, browser_url=e.url)
            for e in catalogue_entries
            for d in (Defn(e.source, e.slug or e.id, e.id),)
        ]
        selections = select_multiple('Select add-ons to install', choices=choices).prompt()
        if selections:
            if confirm('Install selected add-ons?').prompt():
                ctx.invoke(install, addons=selections, replace_folders=False, dry_run=False)
        else:
            click.echo(
                'Nothing was selected; select add-ons with <space>'
                ' and confirm by pressing <enter>.'
            )

    else:
        click.echo('No results found.')


@cli.command('list')
@click.argument('addons', nargs=-1, callback=_with_obj(partial(_parse_uri, raise_invalid=False)))
@click.option(
    '--format',
    '-f',
    'output_format',
    type=_StrEnumChoiceParam(_ListFormat),
    default=_ListFormat.Simple,
    show_default=True,
    help='Change the output format.',
)
@click.pass_obj
def list_installed(
    config_ctx: ConfigBoundCtxProxy, addons: Sequence[Defn], output_format: _ListFormat
) -> None:
    "List installed add-ons."

    with config_ctx.database.connect() as connection:
        where_clause, where_params = _make_pkg_where_clause_and_params(addons)
        pkg_mappings = connection.execute(
            f"""
                SELECT *
                FROM pkg
                {where_clause}
                ORDER BY source, lower(name)
            """,
            where_params,
        ).fetchall()

        def row_mappings_to_pkgs():
            return map(pkg_models.build_pkg_from_row_mapping, repeat(connection), pkg_mappings)

        match output_format:
            case _ListFormat.Json:
                from cattrs.preconf.json import make_converter

                click.echo(
                    make_converter().dumps(list(row_mappings_to_pkgs()), indent=2),
                )

            case _ListFormat.Detailed:

                def format_deps(pkg: pkg_models.Pkg):
                    return (
                        Defn(pkg.source, s or e.id).as_uri()
                        for e in pkg.deps
                        for (s,) in (
                            connection.execute(
                                'SELECT slug FROM pkg WHERE source = :source AND id = :id',
                                {'source': pkg.source, 'id': pkg.id},
                            ).fetchone(),
                        )
                    )

                formatter = click.HelpFormatter(max_width=99)

                for pkg in row_mappings_to_pkgs():
                    with formatter.section(pkg.to_defn().as_uri()):
                        formatter.write_dl(
                            [
                                ('name', pkg.name),
                                ('description', textwrap.shorten(pkg.description, 280)),
                                ('url', pkg.url),
                                ('version', pkg.version),
                                ('date published', pkg.date_published.astimezone().ctime()),
                                ('folders', ', '.join(f.name for f in pkg.folders)),
                                ('dependencies', ', '.join(format_deps(pkg))),
                                (
                                    'options',
                                    '; '.join(
                                        f'{s}={v!r}' for s, v in attrs.asdict(pkg.options).items()
                                    ),
                                ),
                            ]
                        )

                click.echo(formatter.getvalue(), nl=False)

            case _ListFormat.Simple:
                click.echo(
                    ''.join(f'{Defn(p["source"], p["slug"]).as_uri()}\n' for p in pkg_mappings),
                    nl=False,
                )


@cli.command(hidden=True)
@click.argument('addon', callback=_with_obj(partial(_parse_uri, raise_invalid=False)))
@click.pass_context
def info(ctx: click.Context, addon: Defn) -> None:
    "Alias of `list -f detailed`."
    ctx.invoke(list_installed, addons=(addon,), output_format=_ListFormat.Detailed)


@cli.command
@click.argument('addon', callback=_with_obj(partial(_parse_uri, raise_invalid=False)))
@click.pass_obj
def reveal(config_ctx: ConfigBoundCtxProxy, addon: Defn) -> None:
    "Bring an add-on up in your file manager."
    from ._utils.file import reveal_folder

    with config_ctx.database.connect() as connection:
        where_clause, where_params = _make_pkg_where_clause_and_params([addon])
        pkg_folder = connection.execute(
            f"""
                SELECT pkg_folder.name
                FROM pkg
                JOIN pkg_folder ON pkg.source = pkg_folder.pkg_source AND pkg.id = pkg_folder.pkg_id
                {where_clause}
                LIMIT 1
                """,
            where_params,
        ).fetchone()

        if pkg_folder:
            reveal_folder(config_ctx.config.addon_dir / pkg_folder['name'])
        else:
            Report([(addon, R.PkgNotInstalled())]).generate_and_exit()


@cli.command
@click.argument('addons', nargs=-1, callback=_with_obj(partial(_parse_uri, raise_invalid=False)))
@click.option(
    '--convert/--no-convert',
    default=True,
    show_default=True,
    help='Convert HTML and Markdown changelogs to plain text using pandoc. No-op if pandoc is not installed.',
)
@click.option(
    '--remote',
    is_flag=True,
    default=False,
    help='Fetch changelogs from sources.',
)
@click.pass_obj
def view_changelog(
    config_ctx: ConfigBoundCtxProxy, addons: Sequence[Defn], convert: bool, remote: bool
) -> None:
    """View installed and remote add-on changelogs.

    Invoked without arguments, it displays the changelogs of all add-ons
    to have been installed within one minute of the last add-on.

    By default, this command will only retrieve installed add-on changelogs.
    You can reverse this behaviour by passing `--remote`.  With `--remote`,
    you are also able to retrieve changelogs of older versions from sources
    which support the `version_eq` strategy, e.g. `github:foo/bar#version_eq=v1`.
    """

    from ._utils.aio import gather

    MAX_LINES = 100

    def make_converter():
        import shutil

        pandoc = shutil.which('pandoc')
        if pandoc is None:

            def noop_convert(source: str, changelog: str):
                return changelog

            return noop_convert
        else:
            import subprocess

            def real_convert(source: str, changelog: str):
                match config_ctx.resolvers[source].metadata.changelog_format:
                    case ChangelogFormat.Html:
                        pandoc_input_format = 'html'
                    case ChangelogFormat.Markdown:
                        # The "markdown" format will treat a list without a preceding
                        # empty line as a paragraph, which breaks the changelog
                        # of at least one popular add-on.
                        pandoc_input_format = 'commonmark'
                    case _:
                        return changelog

                return subprocess.check_output(
                    [pandoc, '-f', pandoc_input_format, '-t', 'plain'],
                    input=changelog,
                    text=True,
                )

            return real_convert

    def format_combined_changelog_entry(source: str, slug: str, changelog: str):
        lines = changelog.splitlines()
        body = '\n'.join(
            (
                *(f'  {i}' for i in lines[:MAX_LINES]),
                *(('  [...]',) if len(lines) > MAX_LINES else ()),
            )
        )
        return f'{Defn(source, slug).as_uri()}:\n{body}'

    if remote:
        pkgs, _ = pkg_management.bucketise_results(
            run_with_progress(pkg_management.resolve(config_ctx, addons)).items()
        )
        partial_pkgs = [pkg_models.make_db_converter().unstructure(p) for p in pkgs.values()]

    else:
        with config_ctx.database.connect() as connection:
            query = """
                SELECT pkg.source, pkg.slug, pkg.changelog_url
                FROM pkg
            """
            if addons:
                where_clause, query_params = _make_pkg_where_clause_and_params(addons)
                execute = partial(connection.execute, query + where_clause, query_params)
            else:
                query += """
                    JOIN pkg_version_log ON pkg.source = pkg_version_log.pkg_source AND pkg.id = pkg_version_log.pkg_id
                    WHERE pkg_version_log.install_time >= (
                        SELECT datetime(max(pkg_version_log.install_time), '-1 minute')
                        FROM pkg_version_log
                        JOIN pkg ON pkg.source = pkg_version_log.pkg_source AND pkg.id = pkg_version_log.pkg_id
                    )
                """
                execute = partial(connection.execute, query)

            partial_pkgs = execute().fetchall()

    changelogs = run_with_progress(
        gather(
            pkg_management.get_changelog(config_ctx, m['source'], m['changelog_url'])
            for m in partial_pkgs
        )
    )
    if convert:
        do_convert = make_converter()
        changelogs = (do_convert(m['source'], c) for m, c in zip(partial_pkgs, changelogs))

    output = '\n\n'.join(
        format_combined_changelog_entry(m['source'], m['slug'], c)
        for m, c in zip(partial_pkgs, changelogs)
    )
    if output:
        click.echo_via_pager(output)


async def _github_oauth_flow():
    from .github_auth import get_codes, poll_for_access_token
    from .http import init_web_client

    async with init_web_client(None) as web_client:
        codes = await get_codes(web_client)
        click.echo(f'Navigate to {codes["verification_uri"]} and paste the code below:')
        click.echo(f'  {codes["user_code"]}')
        click.echo('Waiting...')
        access_token = await poll_for_access_token(
            web_client, codes['device_code'], codes['interval']
        )
        return access_token


class _EditableConfigOptions(StrEnum):
    AddonDir = 'addon_dir'
    GameFlavour = 'game_flavour'
    AutoUpdateCheck = 'global_config.auto_update_check'
    GithubAccessToken = 'global_config.access_tokens.github'
    CfcoreAccessToken = 'global_config.access_tokens.cfcore'
    WagoAddonsAccessToken = 'global_config.access_tokens.wago_addons'

    path: tuple[str, ...]

    def __new__(cls, value: str) -> Self:
        self = str.__new__(cls, value)
        self._value_ = value
        self.path = tuple(value.split('.'))
        return self


@cli.command
@click.argument(
    'editable-config-values',
    nargs=-1,
    type=_ManyOptionalChoiceValueParam(
        _StrEnumChoiceParam(_EditableConfigOptions),
        value_types={
            _EditableConfigOptions.AutoUpdateCheck: click.types.BoolParamType(),
        },
    ),
)
@click.pass_context
def configure(
    ctx: click.Context,
    editable_config_values: Mapping[_EditableConfigOptions, Any],
) -> _config.ProfileConfig:
    """Configure instawow.

    You can pass configuration keys as arguments to reconfigure an existing
    profile.  Pass a value to suppress the interactive prompt.  For example:

    \b
    * ``configure addon_dir`` will initiate an interactive session
      with autocompletion
    * ``configure "addon_dir=~/foo"` will set ``addon_dir``'s value
      to ``~/foo`` immediately
    """
    import asyncio

    from ._cli_prompts import (
        SKIP,
        AttrsFieldValidator,
        Choice,
        confirm,
        password,
        path,
        select_one,
    )
    from .wow_installations import (
        ADDON_DIR_PARTS,
        Flavour,
        find_installations,
        infer_flavour_from_addon_dir,
    )

    profile = ctx.find_root().params['profile']

    existing_global_config = _config.GlobalConfig.read()

    config_values: dict[str, Any] | None = None
    try:
        config_values = attrs.asdict(_config.ProfileConfig.read(existing_global_config, profile))
    except FileNotFoundError:
        pass
    except Exception:
        _logging.logger.exception('unable to read existing config')

    is_new_profile = config_values is None
    if is_new_profile:
        config_values = {'profile': profile, 'global_config': attrs.asdict(existing_global_config)}
    assert config_values

    editable_config_values = dict(editable_config_values)
    if not editable_config_values:
        default_keys = {
            _EditableConfigOptions.AddonDir,
            _EditableConfigOptions.GameFlavour,
        }
        if existing_global_config.access_tokens.github is None:
            default_keys.add(_EditableConfigOptions.GithubAccessToken)

        editable_config_values = dict.fromkeys(default_keys)

    interactive_editable_config_keys = {k for k, v in editable_config_values.items() if v is None}

    # 0 = Unconfigured
    # 1 = `addon_dir` configured
    # 2 = both `addon_dir` and `game_flavour` configured
    installation_configured = 0
    if (
        is_new_profile
        and {_EditableConfigOptions.AddonDir, _EditableConfigOptions.GameFlavour}
        <= interactive_editable_config_keys
    ):
        known_installations = list(existing_global_config.iter_installations())
        unimported_installations = [
            (k, v and v['flavour'])
            for k, v in find_installations()
            if not any(d.samefile(k) for d in known_installations)
        ]
        if unimported_installations:
            selection = select_one(
                'Installation',
                [Choice(f'{k}  [{v}]', (k, v)) for k, v in unimported_installations],
                can_skip=True,
            ).prompt()
            if selection is not SKIP:
                (installation_path, flavour) = selection

                addon_dir = installation_path.joinpath(*ADDON_DIR_PARTS)
                addon_dir.mkdir(exist_ok=True)

                editable_config_values |= {
                    _EditableConfigOptions.AddonDir: addon_dir,
                    _EditableConfigOptions.GameFlavour: flavour,
                }
                installation_configured = 2 if flavour else 1

    if (
        installation_configured == 0
        and _EditableConfigOptions.AddonDir in interactive_editable_config_keys
    ):
        editable_config_values[_EditableConfigOptions.AddonDir] = path(
            'Add-on directory',
            only_directories=True,
            validator=AttrsFieldValidator(
                attrs.fields(attrs.resolve_types(_config.ProfileConfig)).addon_dir,
                _config.config_converter,
            ),
        ).prompt()

    if (
        installation_configured < 2
        and _EditableConfigOptions.GameFlavour in interactive_editable_config_keys
    ):
        editable_config_values[_EditableConfigOptions.GameFlavour] = select_one(
            'Game flavour',
            [Choice(f, f) for f in Flavour if not f.is_retired],
            initial_value=config_values.get('addon_dir')
            and infer_flavour_from_addon_dir(config_values['addon_dir']),
        ).prompt()

    if _EditableConfigOptions.AutoUpdateCheck in interactive_editable_config_keys:
        editable_config_values[_EditableConfigOptions.AutoUpdateCheck] = confirm(
            'Periodically check for instawow updates?'
        ).prompt()

    if (
        _EditableConfigOptions.GithubAccessToken in interactive_editable_config_keys
        and confirm('Set up GitHub authentication?').prompt()
    ):
        editable_config_values[_EditableConfigOptions.GithubAccessToken] = asyncio.run(
            asyncio.wait_for(_github_oauth_flow(), timeout=60 * 5)
        )

    if _EditableConfigOptions.CfcoreAccessToken in interactive_editable_config_keys:
        click.echo(
            textwrap.fill(
                'An access token is required to use CurseForge. '
                'Log in to https://console.curseforge.com/ to generate an access token.'
            )
        )
        editable_config_values[_EditableConfigOptions.CfcoreAccessToken] = (
            password('CFCore access token:').prompt() or None
        )

    if _EditableConfigOptions.WagoAddonsAccessToken in interactive_editable_config_keys:
        click.echo(
            textwrap.fill(
                'An access token is required to use Wago Addons. '
                'Wago issues tokens to Patreon subscribers above a certain tier. '
                'See https://addons.wago.io/patreon for more information.'
            )
        )
        editable_config_values[_EditableConfigOptions.WagoAddonsAccessToken] = (
            password('Wago Addons access token:').prompt() or None
        )

    for key, value in editable_config_values.items():
        parent = config_values
        for part in key.path[:-1]:
            parent = parent[part]
        parent[key.path[-1]] = value

    config = _config.ProfileConfig.from_values(config_values)
    config.global_config.write()
    config.write()

    click.echo('Configuration written to:')
    click.echo(f'  {config.global_config.config_file}')
    click.echo(f'  {config.config_file}')

    return config


@cli.group('debug')
def _debug_group():
    "Retrieve debugging information."


@_debug_group.command('config')
@click.pass_obj
def debug_config(config_ctx: ConfigBoundCtxProxy) -> None:
    "Print the active configuration."
    import json

    click.echo(json.dumps(config_ctx.config.unstructure_for_display(), indent=2))


@_debug_group.command('sources')
@click.pass_obj
def debug_sources(config_ctx: ConfigBoundCtxProxy) -> None:
    "Print active source metadata."
    from cattrs.preconf.json import make_converter

    json_converter = make_converter()

    click.echo(json_converter.dumps([r.metadata for r in config_ctx.resolvers.values()], indent=2))


@_register_plugin_commands
@cli.group('plugins')
def _plugin_group() -> None:  # pyright: ignore[reportUnusedFunction]
    "Registered plug-ins."


@cli.command(hidden=True)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
def generate_catalogue(start_date: dt.datetime | None) -> None:
    "Generate the master catalogue."
    import asyncio
    import json
    from pathlib import Path

    from ._sources import DEFAULT_RESOLVERS
    from .catalogue.cataloguer import Catalogue, catalogue_converter

    catalogue = asyncio.run(
        Catalogue.collate((r.catalogue for r in DEFAULT_RESOLVERS), start_date)
    )
    catalogue_json = catalogue_converter.unstructure(catalogue)

    catalogue_path = Path(f'base-catalogue-v{catalogue.version}.json').resolve()
    catalogue_path.write_text(
        json.dumps(catalogue_json, indent=2),
        encoding='utf-8',
    )
    catalogue_path.with_suffix(f'.compact{catalogue_path.suffix}').write_text(
        json.dumps(catalogue_json, separators=(',', ':')),
        encoding='utf-8',
    )


main = cli



from __future__ import annotations

import datetime as dt
from functools import lru_cache

import cattrs
from attrs import asdict, field, frozen

from ..definitions import Defn, Strategies, Strategy


def _structure_datetime(value: str, value_type: type):
    return dt.datetime.fromisoformat(value).replace(tzinfo=dt.timezone.utc)


def _unstructure_datetime(value: dt.datetime):
    if not value.tzinfo:
        raise TypeError('`tzinfo` must be set')

    return value.astimezone(dt.timezone.utc).replace(tzinfo=None).isoformat(' ')


@lru_cache(1)
def make_db_converter():
    converter = cattrs.Converter()
    converter.register_structure_hook(dt.datetime, _structure_datetime)
    converter.register_unstructure_hook(dt.datetime, _unstructure_datetime)
    return converter


@frozen(kw_only=True)
class PkgOptions:
    any_flavour: bool
    any_release_type: bool
    version_eq: bool


@frozen(kw_only=True)
class PkgFolder:
    name: str


@frozen(kw_only=True)
class PkgDep:
    id: str


@frozen(kw_only=True)
class PkgLoggedVersion:
    version: str
    install_time: dt.datetime


@frozen(kw_only=True, eq=False)
class Pkg:
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: dt.datetime
    version: str
    changelog_url: str
    options: PkgOptions  # pkg_options
    folders: list[PkgFolder] = field(factory=list)  # pkg_folder
    deps: list[PkgDep] = field(factory=list)  # pkg_dep
    logged_versions: list[PkgLoggedVersion] = field(factory=list)  # pkg_version_log

    def to_defn(self) -> Defn:
        return Defn(
            source=self.source,
            alias=self.slug,
            id=self.id,
            strategies=Strategies(
                dict[Strategy, object](
                    **asdict(
                        self.options,
                        value_serializer=lambda _, a, v: self.version
                        if a.name == Strategy.VersionEq and v is True
                        else v or None,
                    ),
                )
            ),
        )
